# Owner(s): ["oncall: jit"]
# ruff: noqa: F841

import os
import sys
from collections import namedtuple
from typing import Dict, List, NamedTuple, Tuple

import torch
from torch.testing._internal.common_utils import IS_WINDOWS, raise_on_run_directly
from torch.testing._internal.jit_utils import JitTestCase, make_global


# Make the helper files in test/ importable
pytorch_test_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
sys.path.append(pytorch_test_dir)


class TestTyping(JitTestCase):
    def test_dict_in_not_in(self):
        def test_in_dict(x):
            # type: (Dict[str, int]) -> bool
            return "hi" in x

        self.checkScript(test_in_dict, ({"hi": 2, "bye": 3},))
        self.checkScript(test_in_dict, ({"bye": 3},))

        # Check evaluation order
        @torch.jit.script
        def a():
            print("a")
            return 3

        @torch.jit.script
        def b():
            print("b")
            return {3: 2, 4: 1}

        @torch.jit.script
        def fn():
            return a() in b()

        with self.capture_stdout() as captured:
            self.assertTrue(fn())
        if not IS_WINDOWS:
            # no stdout capturing on windows
            self.assertEqual(captured[0], "a\nb\n")

        def test_not_in_dict(a):
            # type: (Dict[str, int]) -> bool
            if "hello" not in a:
                return False
            else:
                return True

        self.checkScript(test_not_in_dict, ({"hello": 1, "world": 2},))
        self.checkScript(test_not_in_dict, ({"world": 2},))

        def test_dict_tensor_key(a, t):
            # type: (Dict[Tensor, int], Tensor) -> bool
            if t in a:
                return True
            else:
                return False

        inp1 = torch.tensor(3)
        inp2 = torch.tensor(5)
        dict_a = {inp1: 1, inp2: 3}
        self.checkScript(test_dict_tensor_key, (dict_a, torch.tensor(4)))
        self.checkScript(test_dict_tensor_key, (dict_a, torch.tensor(3)))
        self.checkScript(test_dict_tensor_key, (dict_a, inp1))
        self.checkScript(test_dict_tensor_key, (dict_a, inp2))

    def test_list_type_refinement_annotation_element_mismatch(self):
        def fn():
            l: List[int] = [1, 2, "foo", 3]
            return l

        with self.assertRaisesRegex(
            RuntimeError,
            "List type annotation"
            r" `List\[int\]` did not match the "
            "types of the given list elements",
        ):
            torch.jit.script(fn)

    def test_dict_type_refinement_annotation_key_mismatch(self):
        def fn():
            l1 = [1, 2, "foo", 3]
            l2 = ["foo", "bar", "baz", "qux"]
            d: Dict[int, str] = dict(zip(l1, l2))
            return d

        with self.assertRaisesRegex(
            RuntimeError,
            "Dicts may only "
            "contain homogeneous keys, but the "
            "type of the first generated key "
            r"was Union\[int, str\]",
        ):
            torch.jit.script(fn)

    def test_dict_type_refinement_annotation_value_mismatch(self):
        def fn():
            l1 = ["foo", "bar", "baz", "qux"]
            l2 = [1, 2, "foo", 3]
            d: Dict[str, int] = dict(zip(l1, l2))
            return d

        with self.assertRaisesRegex(
            RuntimeError,
            "Dict type annotation"
            r" `Dict\[str, int\]` did not match"
            " the type of an actual value type"
            r" `Union\[int, str\]`",
        ):
            torch.jit.script(fn)

    def test_dict_invalid_annotations(self):
        # Check for invalid value type annotation
        def wrong_value_type(dictionary: Dict[str, torch.jit.ScriptModule]):
            return

        with self.assertRaisesRegex(ValueError, "Unknown type annotation"):
            torch.jit.script(wrong_value_type)

        # Check for invalid key type annotation
        def wrong_key_type(dictionary: Dict[torch.jit.ScriptModule, str]):
            return

        with self.assertRaisesRegex(ValueError, "Unknown type annotation"):
            torch.jit.script(wrong_key_type)

        # Check for invalid key and value type annotation
        def wrong_key_value_type(
            dictionary: Dict[torch.jit.ScriptModule, torch.jit.ScriptModule],
        ):
            return

        with self.assertRaisesRegex(ValueError, "Unknown type annotation"):
            torch.jit.script(wrong_key_value_type)

    def test_tuple_specialization(self):
        @torch.jit.script
        def f(t, s):
            # type: (Tuple[Tensor, Tuple[int, Tensor]], str) -> Tensor
            x, t2 = t
            _, y = t2
            return x + y

        t = (
            torch.randn(2, 2),
            (1, torch.randn(2, 2)),
        )
        f(t, "hi")
        graph = f.graph_for(t, "hi")
        input_types = list(next(graph.inputs()).type().elements())
        w = input_types[0]
        self.assertEqual(input_types[0].kind(), "TensorType")
        self.assertEqual(input_types[1].elements()[1].kind(), "TensorType")

    def test_tuple_io(self):
        def stuff(x):
            # type: (Tuple[Tensor, Tensor]) -> Tuple[Tensor, Tensor]
            a, b = x
            return b, a

        a = (torch.rand(3), torch.rand(3))
        self.checkScript(stuff, (a,))

    def test_tuple_keyword(self):
        def bar():
            f = tuple((1, 2))  # noqa: C409
            return f

        self.checkScript(bar, ())

        def foo():
            return tuple(1, 2)

        self.checkScriptRaisesRegex(foo, (), Exception, "1 argument")

        def cant_infer_size():
            return tuple([1, 2, 3])  # noqa: C409

        with self.assertRaisesRegex(Exception, "cannot statically infer the expected"):
            torch.jit.script(cant_infer_size)

    def test_tuple_create_return(self):
        def stuff2(x):
            # type: (int) -> Tuple[Tensor, Tensor]
            a = (torch.ones(x), torch.zeros(x))
            return a

        self.checkScript(stuff2, (3,))

    def test_list_io(self):
        def stuff3(x):
            # type: (List[int]) -> Tuple[Tensor, List[int]]
            return torch.ones(x), x

        self.checkScript(stuff3, ([3, 2],))

    def test_bool_list_io(self):
        @torch.jit.script
        def stuff4(x):
            # type: (List[bool]) -> Tuple[List[bool], List[bool], List[List[bool]]]
            return x, [True, False], [[True]]

        li_1, li_2, li_3 = stuff4([True])
        li_3 = li_3[0]
        for li in [li_1, li_2, li_3]:
            self.assertTrue(type(li[0]) is bool)

    def test_nested_list(self):
        def foo(z):
            # type: (Tuple[int, List[List[int]]]) -> int
            x, y = z
            return y[0][1]

        self.checkScript(foo, ((1, [[1, 2], [3, 4]]),))

    def test_list_sum(self):
        def fn(x: List[int]) -> int:
            return sum(x)

        def fn1(x: List[float]):
            return sum(x)

        def fn2(x: List[bool]):
            return sum(x)

        self.checkScript(fn, ([1, 2, 3],))
        self.checkScript(fn1, ([1.0, 2.0, 3.0],))
        self.checkScript(fn1, ([1, 2.8, 3],))
        self.checkScript(fn2, ([True, False, False],))
        self.checkScript(fn2, ([False, False, False],))
        self.checkScript(fn2, ([0, 1, 1, 0],))

    def test_list_unification(self):
        def fn():
            return [1, None, 2]

        def fn2(x):
            return [torch.ones(2, 2), None, x]

        self.checkScript(fn, [])
        self.checkScript(fn2, (torch.ones(2, 2),))

    # to avoid defining sum_list in multiple tests
    def get_sum_list_fn(self):
        def sum_list(a):
            # type: (List[int]) -> int
            sum = 0
            for i in a:
                sum += i

            return sum

        return sum_list

    def test_sum_list_diff_elms(self):
        self.checkScript(self.get_sum_list_fn(), ([1, 2, 3, 4, 5],))

    def test_sum_list_empty(self):
        self.checkScript(self.get_sum_list_fn(), ([],))

    def test_sum_list_one(self):
        self.checkScript(self.get_sum_list_fn(), ([1],))

    def test_sum_list_literal(self):
        def sum_list():
            # type: () -> int
            sum = 0
            for i in [1, 2, 3, 4, 5]:
                sum += i

            return sum

        self.checkScript(sum_list, ())

    def test_sum_list_wrong_type(self):
        with self.assertRaisesRegex(RuntimeError, "'int' object is not iterable"):

            @torch.jit.script
            def sum_list(a):
                # type: (int) -> int
                sum = 0
                for i in a:  # noqa: T484
                    sum += i

                return sum

            sum_list(1)

    def test_list_iterables(self):
        with self.assertRaisesRegex(
            RuntimeError, "List of iterables is not supported currently"
        ):
            cu = torch.jit.CompilationUnit(
                """
            def list_iterables(x):
                for i, j in [2, 3, 4], [5, 6, 7]:
                    x += i
                    x += j
                return x
            """
            )

    def test_for_in_string(self):
        def test_strings(x):
            # type: (str) -> str
            reverse = ""
            for c in x:
                reverse = c + reverse
            return reverse

        self.checkScript(test_strings, ("hello",))
        self.checkScript(test_strings, ("",))

        def test_list_strings(x):
            # type: (List[str]) -> str
            result = ""
            for sub_str in x:
                result += sub_str
            return result

        self.checkScript(test_list_strings, (["hello", "world"],))
        self.checkScript(test_list_strings, (["hello", " ", "world", ""],))

    def test_for_in_dict(self):
        def test_dicts(x):
            # type: (Dict[str, int]) -> int
            sum = 0
            for key in x:
                sum += x[key]
            return sum

        self.checkScript(test_dicts, ({"a": 1, "b": 2, "c": 3},))

        def test_dict_keys_values(x):
            # type: (Dict[str, int]) -> Tuple[str, int]
            key_str = ""
            sum = 0
            for key in x:
                key_str += key
            for val in x.values():
                sum += val
            return key_str, sum

        self.checkScript(test_dicts, ({"a": 1, "b": 2, "c": 3},))

    def test_for_tuple_unpack(self):
        def for_tuple_unpack(x, y):
            for i, j in [[3, 4], [5, 6], [7, 8]]:
                x += i
                y += j
            return x, y

        self.checkScript(for_tuple_unpack, (torch.tensor(3), torch.tensor(5)))

        def nested_tuple_unpack(x, y):
            # type: (List[int], List[int]) -> int
            sum = 0
            for i, (j, k), v in zip(x, enumerate(x), y):
                sum += i + j + k + v
            return sum

        self.checkScript(nested_tuple_unpack, ([1, 3, 5], [2, 4, 6]))

    def test_dict_comprehension(self):
        def fn():
            return {i: chr(i + 65) for i in range(4)}

        self.checkScript(fn, ())

    def test_dict_comprehension_with_type_annotation(self):
        def fn():
            d: Dict[int, str] = {i: chr(i + 65) for i in range(4)}
            return d

        self.checkScript(fn, ())

        with self.assertRaisesRegex(RuntimeError, ""):
            with self.assertRaisesRegex(
                AssertionError,
                "Expected Dict "
                "type annotation for dict "
                "comprehension, found "
                "Tuple[int, str]",
            ):

                @torch.jit.script
                def fn():
                    d: Tuple[int, str] = {i: chr(i + 65) for i in range(4)}
                    return d

    def test_dict_comprehension_scope(self):
        def comprehension_can_access_outer_scope_variables():
            lst = ["foo", "bar", "baz"]
            return {l: len(l) for l in lst}

        self.checkScript(comprehension_can_access_outer_scope_variables, ())

        with self.assertRaisesRegex(RuntimeError, "undefined value i"):

            @torch.jit.script
            def outer_scope_cannot_access_comprehension_variables():
                d = {i: chr(i + 65) for i in range(4)}
                i = i + 1  # noqa: F821

    def test_for_tuple_assign(self):
        def test_simple_assign(x):
            # type: (Tuple[int, float]) -> float
            sum = 0.0
            for a in x:
                sum += float(a)
            return sum

        self.checkScript(test_simple_assign, ((1, 2.5),))

        def test_tuple_assign(x):
            # type: (Tuple[Tuple[int, int], Tuple[int, int]]) -> int
            sum = 0
            for a in x:
                sum += a[0]
                sum += a[1]
            return sum

        self.checkScript(test_tuple_assign, (((1, 2), (4, 7)),))

        def test_single_starred_lhs(self):
            with self.assertRaisesRegex(
                RuntimeError,
                "A Starred expression may only appear on the lhs within the presence"
                " of another non-starred expression",
            ):
                cu = torch.jit.CompilationUnit(
                    """
                def single_starred_lhs(x):
                    a = (x, x, x)
                    *b, = a
                    return b
                """
                )

    def test_singleton_tuple_unpack(self):
        def foo(a):
            (b,) = (a,)
            return b + 1

        self.checkScript(foo, (torch.rand(3),))

    def test_tuple_assignments(self):
        def var_tuple_assign(x, y):
            # type: (Tuple[Tensor, Tensor], Tensor) -> Tensor
            (a, b), c = x, y
            return a + b + c

        tuple_inputs = (torch.randn(1, 4), torch.randn(3, 4))
        self.checkScript(var_tuple_assign, (tuple_inputs, torch.randn(3, 4)))

        def nested_tuple_assign(x, y, z):
            # type: (int, Tuple[int, Tuple[int, int]], Tuple[int, int]) -> int
            a, (b, (c, d)), (e, f) = x, y, z
            return a + b + c + d + e + f

        self.checkScript(nested_tuple_assign, ((1, (2, (3, 4)), (5, 6))))

        def subscript_tuple_assign(a, x, i):
            # type: (List[int], Tensor, int) -> Tuple[int, Tensor, int]
            a[i], (x[i], b) = 1, (2, 3)
            return a[i] + 1, x + 5, b

        self.checkScript(
            subscript_tuple_assign, ([12, 7, 9, 11], torch.tensor((3, 13, 17)), 0)
        )

        def star_tuple_assign():
            # type: () -> Tuple[int, int, Tuple[int, int], Tuple[int, int]]
            a, (b, *c), *d = 1, (2, 3, 4), 5, 6
            return a, b, c, d

        self.checkScript(star_tuple_assign, ())

        def subscript_tuple_augmented_assign(a):
            # type: (Tuple[int, int]) -> Tuple[int, int]
            a[0] += 1
            return a

        with self.assertRaisesRegex(RuntimeError, "does not support augmented assign"):
            scripted_aug_assign = torch.jit.script(subscript_tuple_augmented_assign)

    def test_multiple_assign(self):
        def test():
            a = b, c = d, f = (1, 1)

            # side effect
            ten = torch.tensor(1)
            ten1 = ten2 = ten.add_(1)

            # ordering
            x = 1
            y = 3
            x, y = y, x + y

            return a, b, c, d, f, ten, ten1, ten2, x, y

        self.checkScript(test, ())

    def test_opt_opt_refinement(self):
        @torch.jit.script
        def test_unify(weight, bias):
            # type: (Optional[int], Optional[int]) -> Optional[int]
            if weight is not None:
                opt = None
            else:
                if bias is not None:
                    opt = 1
                else:
                    opt = None

            return opt

    def test_optional_refinement(self):
        @torch.jit.script
        def test_if_none_assignment(x):
            # type: (Optional[int]) -> int
            if x is None:
                x = 1
            return x + 1

        self.assertEqual(test_if_none_assignment(1), 2)

    def test_optional_conversion(self):
        @torch.jit.script
        def other_fn(x=None):
            # type: (Optional[int]) -> int
            return torch.jit._unwrap_optional(x)

        @torch.jit.script
        def fn(x):
            # type: (int) -> int
            return other_fn(x)

        self.assertEqual(fn(2), 2)

        @torch.jit.script
        def unify_to_optional(x):
            # type: (bool) -> Optional[int]
            if x:
                a = None
            else:
                a = 2
            return a

        self.assertEqual(unify_to_optional(True), None)
        self.assertEqual(unify_to_optional(False), 2)

        @torch.jit.script
        def opt_list(x):
            # type: (Optional[List[float]]) -> int
            return 2

        @torch.jit.script
        def broadcast_opt_list(x):
            # type: (Optional[BroadcastingList2[float]]) -> int
            return 2

        @torch.jit.script
        def opt_list_tuple_caller(x):
            # type: (Tuple[float, float]) -> int
            return opt_list(x) + broadcast_opt_list(x)

        self.assertEqual(opt_list_tuple_caller((2.0, 3.0)), 4)

    def test_optional_tuple(self):
        def fn(x=None):
            # type: (Optional[Tuple[int, int]]) -> Tuple[int, int]
            if x is None:
                new_x = (1, 2)
            else:
                new_x = x
            return new_x

        self.checkScript(fn, ((3, 4),))
        self.checkScript(fn, ())

    def test_namedtuple_redefine(self):
        global _1, _2
        _1 = namedtuple("GoogLeNetOutputs", ["logits", "aux_logits2", "aux_logits1"])
        _2 = namedtuple("GoogLeNetOutputs", ["different"])

        with self.assertRaisesRegex(RuntimeError, r"redefine"):

            @torch.jit.script
            def foo(x, y):
                # type: (_1, _2) -> _1
                return x

    def test_namedtuple_py2(self):
        global _GoogLeNetOutputs  # see [local resolution in python]
        _GoogLeNetOutputs = namedtuple(
            "GoogLeNetOutputs", ["logits", "aux_logits2", "aux_logits1"]
        )

        @torch.jit.script
        def foo(x):
            # type: (_GoogLeNetOutputs) -> _GoogLeNetOutputs
            return x

        vals = torch.rand(3), torch.rand(4), torch.rand(5)
        out = foo(
            _GoogLeNetOutputs(logits=vals[0], aux_logits2=vals[1], aux_logits1=vals[2])
        )
        self.assertEqual(out.logits, vals[0])
        self.assertEqual(out.aux_logits2, vals[1])
        self.assertEqual(out.aux_logits1, vals[2])

    def test_namedtuple_good_error(self):
        global _GoogLeNetOutputs  # see [local resolution in python]
        _GoogLeNetOutputs = namedtuple(
            "GoogLeNetOutputs", ["logits", "aux_logits2", "aux_logits1"]
        )

        @torch.jit.script
        def foo(x):
            # type: (_GoogLeNetOutputs) -> _GoogLeNetOutputs
            return x

        with self.assertRaisesRegex(
            RuntimeError, r"aka NamedTuple\(logits, aux_logits2, aux_logits1\)"
        ):
            out = foo(_GoogLeNetOutputs(logits="3", aux_logits2="4", aux_logits1="5"))

    def test_namedtuple_error_source_attribution(self):
        class _NamedTupleBadMemberType(NamedTuple):
            f1: torch.Tensor
            f2: "ABadForwardRefType"  # noqa: F821

        make_global(_NamedTupleBadMemberType)  # see [local resolution in python]

        def fn(x: _NamedTupleBadMemberType) -> torch.Tensor:
            return x.f1.relu()

        # assert that this has a location associated with the error.
        # note the " +" is regex (i.e. "at least one space")
        with self.assertRaisesRegex(ValueError, "at +File"):
            torch.jit.script(fn)

    def test_inherited_annotations_python_310(self):
        # See #104484
        # In python >=3.10, inspect.get_annotations doesn't always return the same values.
        # Sometimes it will show all annotations; other times it will show only annotations
        # that show in that class, not classes it inherits fro.
        class BaseModule(torch.nn.Module):
            state: List[int]

            def forward(self, x):
                pass

        def do_something_with_list(x: List[int]):
            if x:
                return x[-1]
            return 5

        class Submodule(BaseModule):
            def __init__(self, self_x_value):
                super().__init__()
                self.x = self_x_value
                self.state = []

            def forward(self, x):
                return self.x + x + do_something_with_list(self.state)

        class LowestModule(Submodule):
            def __init__(self) -> None:
                super().__init__(123)

        mod = LowestModule()
        mod2 = LowestModule()
        mod_s = torch.jit.script(mod)
        mod2_s = torch.jit.script(mod2)


if __name__ == "__main__":
    raise_on_run_directly("test/test_jit.py")



# mypy: allow-untyped-defs
import copy
import logging
from typing import Any, Protocol

import torch
from torch._library.utils import parse_namespace
from torch.utils._python_dispatch import _disable_current_modes


log = logging.getLogger(__name__)


class FakeScriptObject:
    def __init__(
        self, wrapped_obj: Any, script_class_name: str, x: torch.ScriptObject | None
    ):
        # Use object.__setattr__ to bypass our custom __setattr__ during initialization
        object.__setattr__(self, "wrapped_obj", wrapped_obj)
        object.__setattr__(self, "script_class_name", script_class_name)

        from torch._library.opaque_object import is_opaque_type

        # We dont want to deepcopy when tracing with opaque objects because
        # if a mutation happens intentionally (Ex. caching in device mesh)
        # then we want it to be recorded on the real object
        real_obj = x
        if not is_opaque_type(type(x)):
            try:
                with _disable_current_modes():
                    real_obj = copy.deepcopy(x)
            except (RuntimeError, TypeError) as e:
                log.warning(  # noqa: G200
                    "Unable to deepcopy the custom object %s due to %s. "
                    "Defaulting to the user given object. This might be "
                    "dangerous as side effects may be directly applied "
                    "to the object.",
                    script_class_name,
                    e,
                )

        object.__setattr__(self, "real_obj", real_obj)

    def __getattribute__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError as e:
            raise AttributeError(
                f"Tried to call __getattr__ with attr '{name}' on a FakeScriptObject, "
                "implying that you are calling this inside of a fake kernel. "
                "The fake kernel should not depend on the contents of the "
                "OpaqueObject at all, so we're erroring out. If this attr is "
                "a method or constant attribute, you can allow this member access by "
                "registering it via `register_opaque_type(members=...)`."
            ) from e

    def __setattr__(self, name, value):
        raise AttributeError(
            f"Tried to call __setattr__ with attr '{name}' on a FakeScriptObject, "
            "implying that you are calling this inside of a fake kernel. "
            "The fake kernel should not depend on the contents of the "
            "OpaqueObject at all, so we're erroring out. If you need this"
            "functionality, consider creating a custom TorchBind Object instead"
            "(but note that this is more difficult)."
        )

    def __getitem__(self, key):
        # This is needed for DeviceMesh support
        return self.real_obj[key]

    def __eq__(self, other):
        if self is other:
            return True
        # Get real_obj without triggering custom __getattribute__
        self_real = object.__getattribute__(self, "real_obj")
        if isinstance(other, FakeScriptObject):
            other_real = object.__getattribute__(other, "real_obj")
            # For reference types, identity check first
            if self_real is other_real:
                return True
            # Fall back to equality check
            return self_real == other_real
        # Compare with the real object directly
        return self_real == other

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        # Use real_obj's hash if available, otherwise use object id
        real_obj = object.__getattribute__(self, "real_obj")
        try:
            return hash(real_obj)
        except TypeError:
            # Object is not hashable, use identity-based hash
            return id(real_obj)

    def __deepcopy__(self, memo: dict[int, Any]) -> "FakeScriptObject":
        if id(self) in memo:
            return memo[id(self)]
        new_obj = FakeScriptObject.__new__(FakeScriptObject)
        memo[id(self)] = new_obj
        object.__setattr__(
            new_obj, "wrapped_obj", copy.deepcopy(self.wrapped_obj, memo)
        )
        object.__setattr__(new_obj, "script_class_name", self.script_class_name)
        # Disable dispatch modes during deepcopy of real_obj and attribute
        # access to prevent tensor operations (e.g. storage cloning, property
        # access on DeviceMesh) from going through proxy tracing or
        # functionalization.
        with _disable_current_modes():
            new_real_obj = copy.deepcopy(self.real_obj, memo)
            object.__setattr__(new_obj, "real_obj", new_real_obj)
            for name, value in self.__dict__.items():
                if name not in ("wrapped_obj", "script_class_name", "real_obj"):
                    if isinstance(value, FakeScriptMethod):
                        object.__setattr__(
                            new_obj,
                            name,
                            FakeScriptMethod(new_obj, value.method_name, value.schema),
                        )
                    else:
                        if hasattr(new_real_obj, name):
                            object.__setattr__(
                                new_obj, name, getattr(new_real_obj, name)
                            )
                        else:
                            object.__setattr__(new_obj, name, value)
        return new_obj


def maybe_unwrap_fake_script_object(obj: Any) -> Any:
    """If obj is a FakeScriptObject, return the underlying real object."""
    if isinstance(obj, FakeScriptObject):
        return obj.real_obj
    return obj


class FakeScriptMethod:
    def __init__(
        self,
        self_fake_obj: FakeScriptObject,
        method_name: str,
        schema: torch.FunctionSchema | None,
    ):
        self.self_fake_obj = self_fake_obj
        self.method_name = method_name
        self.schema = schema

    def __call__(self, *args, **kwargs):
        from torch._higher_order_ops.torchbind import call_torchbind

        return call_torchbind(self.self_fake_obj, self.method_name, *args, **kwargs)


class HasStaticMethodFromReal(Protocol):
    @classmethod
    def from_real(cls, real_obj: torch.ScriptObject):
        pass


class FakeClassRegistry:
    def __init__(self) -> None:
        self._registered_class: dict[str, Any] = {}

    def has_impl(self, full_qualname: str) -> bool:
        return full_qualname in self._registered_class

    def get_impl(self, full_qualname: str) -> Any:
        self._check_registered(full_qualname)
        return self._registered_class[full_qualname]

    def register(self, full_qualname: str, fake_class=None) -> None:
        if self.has_impl(full_qualname):
            log.warning(
                "%s is already registered. Previous fake class is overridden with  %s.",
                full_qualname,
                fake_class,
            )
        self._registered_class[full_qualname] = fake_class

    def deregister(self, full_qualname: str) -> Any:
        if not self.has_impl(full_qualname):
            log.warning(
                "Cannot deregister %s. Please use register_fake_class to register it first."
                " Or do you dereigster it twice?",
                full_qualname,
            )
        else:
            return self._registered_class.pop(full_qualname)

    def clear(self) -> None:
        self._registered_class.clear()

    def _check_registered(self, full_qualname: str) -> None:
        if full_qualname not in self._registered_class:
            raise RuntimeError(
                f"{full_qualname} is not registered. Please use register_fake_class to register it first."
            )


global_fake_class_registry = FakeClassRegistry()


# TODO: add this check at compile time for __obj_flatten__.
def _check_valid_flat_script_obj(flat_x):
    if not isinstance(flat_x, tuple):
        raise RuntimeError("Expect flat x to be a tuple.")

    for tp in flat_x:
        if not isinstance(tp, tuple):
            raise RuntimeError("Expect flat x to be a tuple of tuples.")

        if not len(tp) == 2 or not isinstance(tp[0], str):
            raise RuntimeError(
                "Expect element of flat x to be a tuple of two elements with first element being a string"
            )


def tracing_with_real(x: torch.ScriptObject) -> bool:
    if not hasattr(x, "tracing_mode"):
        return False

    if x.tracing_mode() not in ["real", "fake"]:
        raise AssertionError(
            f"tracing_mode can be either real or fake but got {x.tracing_mode()}"
        )
    return x.tracing_mode() == "real"


def maybe_to_fake_obj(
    fake_mode,
    x: Any,
) -> FakeScriptObject | torch.ScriptObject:
    import torch.utils._pytree as pytree

    # When tracing with real mode, people should implement meta kernels that can
    # handle the case of real script object + fake tensor inputs.
    if tracing_with_real(x):
        return x

    from torch._library.opaque_object import (
        FakeOpaqueObject,
        get_opaque_obj_info,
        get_opaque_type_name,
        is_opaque_type,
        OpaqueTypeStr,
    )

    x_type = type(x)
    if is_opaque_type(x_type):
        type_name = OpaqueTypeStr if x is None else get_opaque_type_name(x_type)
        fake_x_wrapped = FakeScriptObject(FakeOpaqueObject(), type_name, x)

        # Set specified members onto the fake object
        opaque_info = get_opaque_obj_info(x_type)
        if opaque_info is None:
            raise AssertionError(f"opaque_info for type {x_type} must not be None")
        for attr_name in opaque_info.members:
            with _disable_current_modes():
                if not hasattr(x, attr_name):
                    raise TypeError(
                        f"Opaque object of type '{type_name}' was specified to have member "
                        f"'{attr_name}', but this doesn't actually exist in the object."
                    )
                object.__setattr__(fake_x_wrapped, attr_name, getattr(x, attr_name))

        return fake_x_wrapped
    else:
        # x.__obj_flatten__() could be calling some tensor operations inside but we don't
        # want to call these ops in surrounding dispatch modes when executing it.
        # Otherwise, for example, the fake tensor modes will error out when the tensors inside
        # script object execute some operations like clone if allow_non_fake_input flag is set.
        with _disable_current_modes():
            flat_x = x.__obj_flatten__()  # type: ignore[attr-defined]

        _check_valid_flat_script_obj(flat_x)

        with fake_mode:
            from torch._higher_order_ops.utils import _tensor_storage

            storage_map = {
                _tensor_storage(inp): i
                for i, inp in enumerate(flat_x)
                if isinstance(inp, torch.Tensor)
            }
            alias_map = {
                i: storage_map[_tensor_storage(inp)]
                for i, inp in enumerate(flat_x)
                if isinstance(inp, torch.Tensor)
                and storage_map[_tensor_storage(inp)] != i
            }
            if len(alias_map) > 0:
                log.warning(
                    "Detected script object %s has aliasing relationship among its tensors. "
                    "Flattened obj: %s. Aliasing tensor indices: %s. "
                    "This is not supported and may cause unexpected behavior.",
                    x,
                    flat_x,
                    alias_map,
                )

            # This breaks the aliasing relationship among the tensors inside the torchbind object
            # This is bad but since we don't need to preserve the aliasing relationship anyway and
            # we state clearly that aliasing relationship is not preserved in the doc so this might be OK.
            fake_flattened = pytree.tree_map_only(
                torch.Tensor,
                lambda t: torch.empty_strided(
                    t.size(),
                    t.stride(),
                    device=t.device,
                    dtype=t.dtype,
                    requires_grad=t.requires_grad,
                    layout=t.layout,
                ),
                flat_x,
            )

        fake_x = _find_fake_class_for_script_object(x).__obj_unflatten__(fake_flattened)

    fake_x_wrapped = FakeScriptObject(fake_x, x._type().qualified_name(), x)  # type: ignore[attr-defined]

    for name in x._method_names():  # type: ignore[attr-defined]
        attr = getattr(fake_x, name, None)
        if attr is not None:
            if not callable(attr):
                raise RuntimeError(f"Expect {name} to be a callable but got {attr}.")

            real_attr = getattr(x, name)  # type: ignore[attr-defined]

            # real attr sometimes is not torch.ScriptMethod thus doesn't have schema e.g. __init___ or __eq__
            method_schema: torch.FunctionSchema | None = None
            if isinstance(real_attr, torch.ScriptMethod):
                method_schema = real_attr.schema  # type: ignore[attr-defined]

            # Bypasses our custom setattr function
            object.__setattr__(
                fake_x_wrapped,
                name,
                FakeScriptMethod(fake_x_wrapped, name, method_schema),
            )
        else:
            override_skip_list = {"__obj_flatten__", "__getstate__", "__setstate__"}
            if name not in override_skip_list:
                log.warning("fake object of %s doesn't implement method %s.", x, name)
    return fake_x_wrapped


def register_fake_class(qualname, fake_class: HasStaticMethodFromReal | None = None):
    r"""Register a fake implementation for this class.

    It's in the same spirit of registering a fake implementation for
    an operator but with the difference that it
    associates a fake class with the original torch bind class (registered
    with torch::class_). In this way, torch.compile can handle them properly
    in components such as Dynamo and AOTAutograd.

    This API may be used as a decorator (see example). For the fake class, users
    are required to provide a from_real classmethod that takes a real object and
    returns an instance of the fake class. All tensors in the fake object should also
    be properly fakified with to_fake_tensor() in from_real.


    Examples:
        # For a custom class Foo defined in test_custom_class_registration.cpp:

        TORCH_LIBRARY(_TorchScriptTesting, m) {
          m.class_<TensorQueue>("_TensorQueue")
            .def(torch::init<at::Tensor>())
            .def("push", &TensorQueue::push)
            .def("pop", &TensorQueue::pop)
            .def("top", &TensorQueue::top)
            .def("size", &TensorQueue::size)
            .def("clone_queue", &TensorQueue::clone_queue)
            .def("__obj_flatten__", &TensorQueue::__obj_flatten__)
            .def_pickle(
                // __getstate__
                [](const c10::intrusive_ptr<TensorQueue>& self)
                    -> c10::Dict<std::string, at::Tensor> {
                  return self->serialize();
                },
                // __setstate__
                [](c10::Dict<std::string, at::Tensor> data)
                    -> c10::intrusive_ptr<TensorQueue> {
                  return c10::make_intrusive<TensorQueue>(std::move(data));
                });
            };
        # We could register a fake class FakeTensorQueue in Python as follows:
        import torch

        @torch._library.register_fake_class("_TorchScriptTesting::_TensorQueue")
        class FakeTensorQueue:
            def __init__(self, queue):
                self.queue = queue

            @classmethod
            def __obj_unflatten__(cls, flattened_ctx):
                return cls(**dict(ctx))

            def push(self, x):
                self.queue.append(x)

            def pop(self):
                return self.queue.pop(0)

            def size(self):
                return len(self.queue)

    In this example, the original TensorQeue need to add a __obj_flatten__ method
    to the class TensorQueue and the flattened result is passed into FakeTensorQueue's
    __obj_unflatten__ as inputs to create a fake class. This protocol allows pytorch to look
    at the contents of the script object and properly handle them in the subsystems
    like dynamo, aot_aotugrad or more.
    """

    def inner(fake_class: HasStaticMethodFromReal):
        ns, name = parse_namespace(qualname)

        # This also checks whether the referred torch::class_ exists.
        torch._C._get_custom_class_python_wrapper(ns, name)

        from_method = getattr(fake_class, _CONVERT_FROM_REAL_NAME, None)
        if not from_method:
            raise RuntimeError(
                f"{fake_class} doesn't define a classmethod {_CONVERT_FROM_REAL_NAME}."
            )

        if not isinstance(fake_class.__dict__[_CONVERT_FROM_REAL_NAME], classmethod):
            raise RuntimeError(
                f"{_CONVERT_FROM_REAL_NAME} method is not a classmethod."
            )

        global_fake_class_registry.register(_full_qual_class_name(qualname), fake_class)
        return fake_class

    if fake_class is None:
        return inner
    return inner(fake_class)


def deregister_fake_class(qualname):
    return global_fake_class_registry.deregister(_full_qual_class_name(qualname))


def has_fake_class(full_qualname) -> bool:
    return global_fake_class_registry.has_impl(full_qualname)


def find_fake_class(full_qualname) -> Any | None:
    if not has_fake_class(full_qualname):
        return None
    return global_fake_class_registry.get_impl(full_qualname)


def _full_qual_class_name(qualname: str) -> str:
    ns, name = parse_namespace(qualname)
    return "__torch__.torch.classes." + ns + "." + name


def _is_script_object(obj: Any) -> bool:
    return isinstance(
        obj, torch.ScriptObject
    ) and obj._type().qualified_name().startswith(  # type: ignore[attr-defined]
        "__torch__.torch.classes"
    )


# Return the namespace and class name from fully qualified name.
def _ns_and_class_name(full_qualname: str) -> tuple[str, str]:
    splits = full_qualname.split(".")
    if len(splits) != 5:
        raise AssertionError(f"Could not split {full_qualname=}, expected 5 parts")
    _torch, _torch_ns, _classes, ns, class_name = splits
    return ns, class_name


def _find_fake_class_for_script_object(x: torch.ScriptObject) -> Any:
    full_qualname = x._type().qualified_name()  # type: ignore[attr-defined]
    ns, class_name = _ns_and_class_name(full_qualname)
    fake_class = find_fake_class(full_qualname)
    if fake_class is None:
        raise RuntimeError(
            f" ScriptObject's {full_qualname} haven't registered a fake class."
            f" Please use register_fake_class({ns}::{class_name}) to annotate a fake class for the script obj."
            f" Specifically, create a python class that implements a fake version for all the methods"
            f" that're used in the program and put annotated class in the program e.g. after loading the library."
            f" The fake methods can be written in the same way as a meta kernel for an operator but need to additionally"
            f" simulate the object's states. Be sure to add a {_CONVERT_FROM_REAL_NAME} classmethod"
            f" to enable creating a fake obj from a real one."
        )
    return fake_class


_CONVERT_FROM_REAL_NAME = "__obj_unflatten__"


def _fake_obj_from_real(fake_mode, x) -> Any:
    fake_class = _find_fake_class_for_script_object(x)

    from_real_method = getattr(fake_class, _CONVERT_FROM_REAL_NAME, None)
    if not from_real_method:
        raise RuntimeError(
            f"{fake_class} must define a classmethod {_CONVERT_FROM_REAL_NAME}"
            f" that converts the real object to the fake object."
        )

    # from_real defined by user need the ctx to fakify the tensor states.
    ctx = torch._library.fake_impl.FakeImplCtx(fake_mode, None)
    with torch._library.fake_impl.set_ctx_getter(lambda: ctx):
        return fake_class.from_real(x)



"""
Dynamo Profiler - tracks where Dynamo spends time during compilation.

This module provides profiling functionality for Dynamo tracing, showing per-function
cumtime (inclusive) and tottime (exclusive) in a cProfile-compatible format.
The output can be visualized with tools like snakeviz.

Usage:
    # Enable via config (prints pstats output):
    torch._dynamo.config.dynamo_profiler = True

    # Or save to file for snakeviz:
    torch._dynamo.config.dynamo_profiler = "/tmp/dynamo.prof"
    # Then: snakeviz /tmp/dynamo.prof
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, TYPE_CHECKING


if TYPE_CHECKING:
    import pstats


@dataclass
class FunctionTraceTiming:
    """
    Timing data for a single inlined function trace.

    Follows cProfile conventions:
    - cumtime: total time in function including all subcalls (inclusive)
    - tottime: time in function excluding subcalls (exclusive)
    - caller info: who called this function (for building call graph)
    """

    # Function identification
    func_name: str
    filename: str
    firstlineno: int
    # Timing data (in nanoseconds) - cProfile-style
    cumtime_ns: int  # Inclusive time (includes subcalls)
    tottime_ns: int  # Exclusive time (excludes subcalls)
    # Code stats (for comparing tracing overhead vs function complexity)
    bytecode_count: int
    # Nesting depth when this function was traced
    inline_depth: int
    # Caller information (for building call graph edges)
    caller_func_name: str | None = None
    caller_filename: str | None = None
    caller_firstlineno: int | None = None
    # Whether this is a primitive (non-recursive) call
    # A call is primitive if the function doesn't appear anywhere in the call stack
    is_primitive_call: bool = True
    # Full call stack at the time of this call (for proper snakeviz drill-down)
    # Each entry is (func_name, filename, firstlineno)
    call_stack: tuple[tuple[str, str, int], ...] = ()

    # Backwards compatibility alias
    @property
    def trace_time_ns(self) -> int:
        return self.cumtime_ns

    @property
    def trace_time_ms(self) -> float:
        return self.cumtime_ns / 1e6

    @property
    def cumtime_ms(self) -> float:
        return self.cumtime_ns / 1e6

    @property
    def tottime_ms(self) -> float:
        return self.tottime_ns / 1e6

    @property
    def caller_key(self) -> tuple[str, int, str] | None:
        """Return caller as a pstats-compatible key tuple."""
        if self.caller_func_name is not None:
            return (
                self.caller_filename or "",
                self.caller_firstlineno or 0,
                self.caller_func_name,
            )
        return None

    @property
    def func_key(self) -> tuple[str, int, str]:
        """Return this function as a pstats-compatible key tuple."""
        return (self.filename, self.firstlineno, self.func_name)

    def __repr__(self) -> str:
        return (
            f"FunctionTraceTiming({self.func_name} at {self.filename}:{self.firstlineno}, "
            f"cumtime={self.cumtime_ms:.2f}ms, tottime={self.tottime_ms:.2f}ms, "
            f"bytecode={self.bytecode_count}, depth={self.inline_depth})"
        )


@dataclass
class ProfilerStackEntry:
    """Stack entry for tracking function timing in the Dynamo profiler."""

    func_name: str
    filename: str
    firstlineno: int
    start_time_ns: int
    child_time_ns: int  # Accumulated time spent in traced children
    is_primitive_call: bool = True  # Whether this is a non-recursive call


class DynamoProfilerState:
    """State for Dynamo profiler tracking function trace timings."""

    def __init__(self) -> None:
        self.timings: list[FunctionTraceTiming] = []
        self.stack: list[ProfilerStackEntry] = []

    def record_timing(self, timing: FunctionTraceTiming) -> None:
        """Record timing data for a traced function."""
        self.timings.append(timing)

    def get_timings(self) -> list[FunctionTraceTiming]:
        """Get all recorded timings."""
        return self.timings

    def push(
        self, func_name: str, filename: str, firstlineno: int, start_time_ns: int
    ) -> None:
        """Push a new entry onto the timing stack."""
        # Check if this function already exists in the stack (indirect recursion)
        is_primitive = not any(
            entry.func_name == func_name
            and entry.filename == filename
            and entry.firstlineno == firstlineno
            for entry in self.stack
        )
        self.stack.append(
            ProfilerStackEntry(
                func_name=func_name,
                filename=filename,
                firstlineno=firstlineno,
                start_time_ns=start_time_ns,
                child_time_ns=0,
                is_primitive_call=is_primitive,
            )
        )

    def pop(self) -> ProfilerStackEntry | None:
        """Pop the top entry from the timing stack."""
        if self.stack:
            return self.stack.pop()
        return None

    def add_child_time(self, child_cumtime_ns: int) -> None:
        """Add the child's cumulative time to the parent's child_time accumulator."""
        if self.stack:
            self.stack[-1].child_time_ns += child_cumtime_ns

    def get_current_caller(self) -> tuple[str, str, int] | None:
        """Get the current caller (top of stack) as (func_name, filename, firstlineno)."""
        if self.stack:
            entry = self.stack[-1]
            return (entry.func_name, entry.filename, entry.firstlineno)
        return None

    def get_call_stack(self) -> tuple[tuple[str, str, int], ...]:
        """Get the full current call stack as tuple of (func_name, filename, firstlineno)."""
        return tuple(
            (entry.func_name, entry.filename, entry.firstlineno) for entry in self.stack
        )

    def generate_pstats(
        self, output_file: str | None = None, print_raw: bool = False
    ) -> pstats.Stats:
        """Generate pstats.Stats object from recorded timings.

        Args:
            output_file: Optional file path to save the stats.
            print_raw: If True, print raw aggregate timings before returning.
        """
        import cProfile
        import io
        import logging
        import pstats

        log = logging.getLogger(__name__)

        # Aggregate by (filename, lineno, func_name)
        aggregated: dict[tuple[str, int, str], dict[str, Any]] = {}
        # caller_edges[callee_key][caller_key] -> edge stats
        caller_edges: dict[
            tuple[str, int, str], dict[tuple[str, int, str], dict[str, Any]]
        ] = {}

        for t in self.timings:
            key = (t.filename, t.firstlineno, t.func_name)

            if key not in aggregated:
                aggregated[key] = {
                    "ncalls": 0,
                    "pcalls": 0,
                    "tottime": 0.0,
                    "cumtime": 0.0,
                }
                caller_edges[key] = {}

            agg = aggregated[key]
            agg["ncalls"] += 1
            agg["tottime"] += t.tottime_ns / 1e9

            if t.is_primitive_call:
                agg["pcalls"] += 1
                agg["cumtime"] += t.cumtime_ns / 1e9

            # Build caller edge
            if t.caller_filename is not None:
                caller_key = (
                    t.caller_filename,
                    t.caller_firstlineno or 0,
                    t.caller_func_name or "",
                )
                if caller_key not in caller_edges[key]:
                    caller_edges[key][caller_key] = {
                        "ncalls": 0,
                        "pcalls": 0,
                        "tottime": 0.0,
                        "cumtime": 0.0,
                    }
                edge = caller_edges[key][caller_key]
                edge["ncalls"] += 1
                edge["tottime"] += t.tottime_ns / 1e9
                # Always add cumtime to edges for visualization (gprof2dot)
                # Function-level cumtime is already correct (only primitive calls)
                edge["cumtime"] += t.cumtime_ns / 1e9
                if t.is_primitive_call:
                    edge["pcalls"] += 1

        if print_raw:
            sorted_items = sorted(
                aggregated.items(), key=lambda x: x[1]["cumtime"], reverse=True
            )
            print("\n=== Aggregate Timings (raw) ===")
            print(
                f"{'ncalls':>8} {'pcalls':>8} {'tottime':>12} {'cumtime':>12}  function"
            )
            print("-" * 80)
            total_cumtime = 0.0
            total_tottime = 0.0
            for (filename, lineno, func_name), agg in sorted_items:
                ncalls = agg["ncalls"]
                pcalls = agg["pcalls"]
                tottime = agg["tottime"] * 1000  # convert to ms
                cumtime = agg["cumtime"] * 1000
                total_cumtime += cumtime
                total_tottime += tottime
                short_file = filename.split("/")[-1] if "/" in filename else filename
                print(
                    f"{ncalls:>8} {pcalls:>8} {tottime:>10.2f}ms {cumtime:>10.2f}ms  "
                    f"{func_name} ({short_file}:{lineno})"
                )
            print("-" * 80)
            print(
                f"Total timings: {len(self.timings)}, unique functions: {len(aggregated)}"
            )
            print(
                f"Sum tottime: {total_tottime:.2f}ms, Sum cumtime: {total_cumtime:.2f}ms"
            )

        # Ensure caller-only functions have a top-level entry.
        # gprof2dot expects every function referenced as a caller to also
        # exist as a top-level entry in the stats dict with timing data.
        for key in list(caller_edges.keys()):
            for caller_key in caller_edges[key]:
                if caller_key not in aggregated:
                    aggregated[caller_key] = {
                        "ncalls": 0,
                        "pcalls": 0,
                        "tottime": 0.0,
                        "cumtime": 0.0,
                    }
                    caller_edges[caller_key] = {}

        # Build the stats dict in pstats format
        stats_dict: dict[
            tuple[str, int, str], tuple[int, int, float, float, dict[Any, Any]]
        ] = {}

        for key, agg in aggregated.items():
            callers: dict[tuple[str, int, str], tuple[int, int, float, float]] = {}
            for caller_key, edge in caller_edges[key].items():
                callers[caller_key] = (
                    edge["ncalls"],
                    edge["pcalls"],
                    edge["tottime"],
                    edge["cumtime"],
                )

            stats_dict[key] = (
                agg["pcalls"],
                agg["ncalls"],
                agg["tottime"],
                agg["cumtime"],
                callers,
            )

        # Create a pstats.Stats object
        dummy_profile = cProfile.Profile()
        dummy_profile.enable()
        dummy_profile.disable()
        stats = pstats.Stats(dummy_profile, stream=io.StringIO())

        stats.stats = stats_dict  # type: ignore[attr-defined]
        stats.total_calls = sum(s[1] for s in stats_dict.values())  # type: ignore[attr-defined]
        stats.prim_calls = sum(s[0] for s in stats_dict.values())  # type: ignore[attr-defined]
        stats.total_tt = sum(s[2] for s in stats_dict.values())  # type: ignore[attr-defined]

        if output_file:
            stats.dump_stats(output_file)
            log.info(
                "Saved pstats to %s. Visualize with: snakeviz %s",
                output_file,
                output_file,
            )

        return stats

    def generate_svg(
        self, profile_file: str, svg_file: str | None = None
    ) -> str | None:
        """Generate an SVG call graph from a profile file using gprof2dot and graphviz.

        Args:
            profile_file: Path to the pstats profile file.
            svg_file: Optional path for the output SVG. If not provided, uses
                profile_file with .svg extension.

        Returns:
            Path to the generated SVG file, or None if generation failed.
        """
        import os
        import shutil
        import subprocess

        if not shutil.which("gprof2dot"):
            print("gprof2dot not found. Install with: pip install gprof2dot")
            return None

        if not shutil.which("dot"):
            print("graphviz 'dot' not found. Install graphviz package.")
            return None

        if svg_file is None:
            svg_file = profile_file.rsplit(".", 1)[0] + ".svg"

        try:
            # gprof2dot -f pstats profile.prof | dot -Tsvg -o profile.svg
            gprof2dot = subprocess.Popen(
                [
                    "gprof2dot",
                    "-f",
                    "pstats",
                    "--node-label=total-time-percentage",
                    "--node-label=self-time-percentage",
                    "--node-label=total-time",
                    profile_file,
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            dot = subprocess.Popen(
                ["dot", "-Tsvg", "-o", svg_file],
                stdin=gprof2dot.stdout,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            gprof2dot.stdout.close()  # type: ignore[union-attr]
            _, dot_err = dot.communicate()
            _, gprof2dot_err = gprof2dot.communicate()

            if gprof2dot.returncode != 0:
                print(
                    f"gprof2dot failed: {gprof2dot_err.decode()}"  # noqa: B950
                )
                return None

            if dot.returncode != 0:
                print(f"graphviz dot failed: {dot_err.decode()}")
                return None

            if not os.path.isfile(svg_file):
                print(f"SVG file was not created: {svg_file}")
                return None

            print(f"SVG call graph saved to: {svg_file}")
            return svg_file

        except Exception as e:
            print(f"Failed to generate SVG: {e}")
            return None

    def dump_stats(
        self, output_file: str | None = None, generate_svg: bool = True
    ) -> None:
        """Print profiler stats to stdout and optionally save to file.

        Args:
            output_file: Optional path to save the pstats profile.
            generate_svg: If True and output_file is provided, also generate an SVG
                call graph using gprof2dot and graphviz.
        """
        import sys

        if not self.timings:
            return

        stats = self.generate_pstats(output_file, print_raw=True)
        print("\n=== Dynamo Profiler (pstats) ===")
        stats.stream = sys.stdout  # type: ignore[attr-defined]
        stats.sort_stats("cumulative").print_stats()

        if output_file:
            print(f"\nProfile saved to: {output_file}")
            print(f"Visualize with: snakeviz {output_file}")

            if generate_svg:
                self.generate_svg(output_file)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional

import torch
from torch.library import register_fake


torch.library.define(
    "blackwell_fmha::fmha_fwd",
    "(Tensor q, Tensor k, Tensor v, Tensor? cu_seqlens_q, Tensor? cu_seqlens_k, int? max_seq_len_q, int? max_seq_len_k, float? softmax_scale, bool? causal, Tensor? seqlen_kv, Tensor? page_table, int seqlen_k=-1, int window_size_left=-1, int window_size_right=-1, bool bottom_right=True) -> (Tensor, Tensor)",
    tags=torch.Tag.pt2_compliant_tag,
)

torch.library.define(
    "blackwell_fmha::fmha_bwd",
    "(Tensor dout, Tensor q, Tensor k, Tensor v, Tensor out, Tensor softmax_lse, Tensor? cu_seqlens_q, Tensor? cu_seqlens_k, int? max_seq_len_q, int? max_seq_len_k, float? softmax_scale, bool? causal, int window_size_left=-1, int window_size_right=-1, bool bottom_right=True, bool deterministic=False) -> (Tensor, Tensor, Tensor)",
    tags=torch.Tag.pt2_compliant_tag,
)


@torch.library.impl("blackwell_fmha::fmha_fwd", "cuda")
def custom_op_fmha(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    cu_seqlens_q: Optional[torch.Tensor] = None,
    cu_seqlens_k: Optional[torch.Tensor] = None,
    max_seq_len_q: Optional[int] = None,
    max_seq_len_k: Optional[int] = None,
    softmax_scale: Optional[float] = None,
    causal: bool = False,
    seqlen_kv: Optional[torch.Tensor] = None,
    page_table: Optional[torch.Tensor] = None,
    seqlen_k: Optional[int] = None,
    window_size_left: int = -1,
    window_size_right: int = -1,
    bottom_right: bool = True,
) -> tuple[torch.Tensor, torch.Tensor]:
    assert q.is_contiguous(), "q is not contiguous"
    assert k.is_contiguous(), "k is not contiguous"
    assert v.is_contiguous(), "v is not contiguous"
    assert q.is_cuda, "q must be on GPU"
    assert k.is_cuda, "k must be on GPU"
    assert v.is_cuda, "v must be on GPU"

    return torch.ops.mslk.fmha_fwd(
        q,
        k,
        v,
        cu_seqlens_q=cu_seqlens_q,
        cu_seqlens_k=cu_seqlens_k,
        max_seq_len_q=max_seq_len_q,
        max_seq_len_k=max_seq_len_k,
        softmax_scale=softmax_scale,
        causal=causal,
        seqlen_kv=seqlen_kv,
        page_table=page_table,
        seqlen_k=seqlen_k,
        window_size_left=window_size_left,
        window_size_right=window_size_right,
        bottom_right=bottom_right,
    )


@register_fake("blackwell_fmha::fmha_fwd")
def fmha_fwd_meta(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    cu_seqlens_q: Optional[torch.Tensor] = None,
    cu_seqlens_k: Optional[torch.Tensor] = None,
    max_seq_len_q: Optional[int] = None,
    max_seq_len_k: Optional[int] = None,
    softmax_scale: Optional[float] = None,
    causal: bool = False,
    seqlen_kv: Optional[torch.Tensor] = None,
    page_table: Optional[torch.Tensor] = None,
    seqlen_k: Optional[int] = None,
    window_size_left: int = -1,
    window_size_right: int = -1,
    bottom_right: bool = True,
):
    if q.dtype == torch.float16:
        out_dtype = torch.float16
    elif q.dtype == torch.bfloat16:
        out_dtype = torch.bfloat16
    elif q.dtype == torch.float8_e4m3fn:
        # Output is BF16 when input is FP8
        out_dtype = torch.bfloat16
    else:
        raise RuntimeError(f"Unsupported dtype for q: {q.dtype}")

    kIsVarlen = max_seq_len_q is not None
    if kIsVarlen:
        assert cu_seqlens_q is not None
        SQ = q.shape[0]
        H_Q = q.shape[1]
        B = cu_seqlens_q.shape[0] - 1
    else:
        SQ = q.shape[1]
        H_Q = q.shape[2]
        B = q.shape[0]
    device = q.device
    options2 = {"dtype": torch.float32, "device": device}
    if kIsVarlen:
        assert max_seq_len_q is not None
        out = torch.empty_like(q, dtype=out_dtype)
        size = out.size()
        stride = out.stride()
        storage_offset = q.shape[-1] * max_seq_len_q * H_Q  # example scalar offset
        out1 = torch.as_strided(
            out, size=size, stride=stride, storage_offset=storage_offset
        )
    else:
        out1 = torch.empty_like(q, dtype=out_dtype)

    if kIsVarlen:
        out2 = torch.empty((1, H_Q, SQ), **options2)  # type: ignore
    else:
        out2 = torch.empty((B, H_Q, SQ), **options2)  # type: ignore
    return out1, out2


@torch.library.impl("blackwell_fmha::fmha_bwd", "cuda")
def custom_op_fmha_bwd(
    dOutput: torch.Tensor,
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    output: torch.Tensor,
    softmax_lse: torch.Tensor,
    cu_seqlens_q: Optional[torch.Tensor] = None,
    cu_seqlens_k: Optional[torch.Tensor] = None,
    max_seq_len_q: Optional[int] = None,
    max_seq_len_k: Optional[int] = None,
    softmax_scale: Optional[float] = None,
    causal: bool = False,
    window_size_left: int = -1,
    window_size_right: int = -1,
    bottom_right: bool = True,
    deterministic: bool = False,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    return torch.ops.mslk.fmha_bwd(
        dOutput,
        query,
        key,
        value,
        output,
        softmax_lse,
        cu_seqlens_q=cu_seqlens_q,
        cu_seqlens_k=cu_seqlens_k,
        max_seq_len_q=max_seq_len_q,
        max_seq_len_k=max_seq_len_k,
        softmax_scale=softmax_scale,
        causal=causal,
        window_size_left=window_size_left,
        window_size_right=window_size_right,
        bottom_right=bottom_right,
        deterministic=deterministic,
    )


@register_fake("blackwell_fmha::fmha_bwd")
def fmha_bwd_meta(
    dOutput: torch.Tensor,
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    output: torch.Tensor,
    softmax_lse: torch.Tensor,
    cu_seqlens_q: Optional[torch.Tensor] = None,
    cu_seqlens_k: Optional[torch.Tensor] = None,
    max_seq_len_q: Optional[int] = None,
    max_seq_len_k: Optional[int] = None,
    softmax_scale: Optional[float] = None,
    causal: bool = False,
    window_size_left: int = -1,
    window_size_right: int = -1,
    bottom_right: bool = True,
    deterministic: bool = False,
):
    return (
        torch.empty_like(query),
        torch.empty_like(key),
        torch.empty_like(value),
    )


def _backward(ctx, *grad):
    if ctx.is_gen:
        # For gen case, no backward pass is needed (generation is inference only)
        raise RuntimeError("Backward pass is not supported for generation phase (sq=1)")
    q, k, v, out, softmax_lse = ctx.saved_tensors
    if not grad[0].is_contiguous():
        grad0 = grad[0].contiguous()
    else:
        grad0 = grad[0]
    if not softmax_lse.is_contiguous:
        softmax_lse = softmax_lse.contiguous()
    if not out.is_contiguous:
        out = out.contiguous()
    if not q.is_contiguous:
        q = q.contiguous()
    if not k.is_contiguous:
        k = k.contiguous()

    if not softmax_lse.is_contiguous:
        softmax_lse = softmax_lse.contiguous()
    if not out.is_contiguous:
        out = out.contiguous()
    if not q.is_contiguous:
        q = q.contiguous()
    if not k.is_contiguous:
        k = k.contiguous()

    dq, dk, dv = torch.ops.blackwell_fmha.fmha_bwd(
        grad0,
        q,
        k,
        v,
        out,
        softmax_lse,
        ctx.cu_seqlens_q,
        ctx.cu_seqlens_k,
        ctx.max_seq_len_q,
        ctx.max_seq_len_k,
        ctx.softmax_scale,
        ctx.causal,
        ctx.window_size_left,
        ctx.window_size_right,
        ctx.bottom_right,
        ctx.deterministic,
    )
    return (
        dq,
        dk,
        dv,
        None,
        None,
        None,
        None,
        None,
        None,
        None,
        None,
        None,
        None,
        None,
        None,
    )


def _setup_context(ctx, inputs, output):
    (
        q,
        k,
        v,
        cu_seqlens_q,
        cu_seqlens_k,
        max_seq_len_q,
        max_seq_len_k,
        softmax_scale,
        causal,
        seqlen_kv,
        page_table,
        seqlen_k,
        window_size_left,
        window_size_right,
        bottom_right,
    ) = inputs
    (out, softmax_lse) = output
    ctx.save_for_backward(q, k, v, out, softmax_lse)
    ctx.softmax_scale = softmax_scale
    ctx.causal = causal
    ctx.max_seq_len_q = max_seq_len_q
    ctx.max_seq_len_k = max_seq_len_k
    ctx.cu_seqlens_q = cu_seqlens_q
    ctx.cu_seqlens_k = cu_seqlens_k
    ctx.window_size_left = window_size_left
    ctx.window_size_right = window_size_right
    ctx.bottom_right = bottom_right
    ctx.deterministic = False  # Set default value
    ctx.is_gen = False


# This code adds training support for the operator. You must provide us
# the backward formula for the operator and a `setup_context` function
# to save values to be used in the backward.
torch.library.register_autograd(
    "blackwell_fmha::fmha_fwd", _backward, setup_context=_setup_context
)


def cutlass_blackwell_fmha_custom_op(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    softmax_scale: float | None = None,
    causal: bool = False,
    cu_seqlens_q: torch.Tensor | None = None,
    cu_seqlens_k: torch.Tensor | None = None,
    max_seq_len_q: int | None = None,
    max_seq_len_k: int | None = None,
    seqlen_kv: torch.Tensor | None = None,
    page_table: torch.Tensor | None = None,
    seqlen_k: int | None = -1,
    window_size_left: int | None = -1,
    window_size_right: int | None = -1,
    bottom_right: bool | None = True,
):
    return torch.ops.blackwell_fmha.fmha_fwd(
        q=q,
        k=k,
        v=v,
        cu_seqlens_q=cu_seqlens_q,
        cu_seqlens_k=cu_seqlens_k,
        max_seq_len_q=max_seq_len_q,
        max_seq_len_k=max_seq_len_k,
        softmax_scale=softmax_scale,
        causal=causal,
        seqlen_kv=seqlen_kv,
        page_table=page_table,
        seqlen_k=seqlen_k,
        window_size_left=window_size_left,
        window_size_right=window_size_right,
        bottom_right=bottom_right,
    )[0]



import functools
import logging
from collections.abc import Callable, Sequence
from typing import TypeAlias

import torch
from torch.fx import Node

from .common import CantChunk
from .core import get_chunking_meta, get_chunking_metas, log, update_chunking_meta
from .utils import (
    format_node_with_chunking_meta,
    get_args_of_node_type,
    get_scale_by_from_metas,
    get_scale_by_from_node,
)


aten = torch.ops.aten
prims = torch.ops.prims

_HandlerType: TypeAlias = Callable[[Node], bool]
propagate_rules: dict[torch._ops.OpOverload, _HandlerType] = {}


def _register_propagate_rule(
    aten_op: torch._ops.OpOverload | Sequence[torch._ops.OpOverload],
    handler: _HandlerType,
) -> _HandlerType:
    if not isinstance(aten_op, (list, tuple)):
        aten_op = [aten_op]  # type: ignore[assignment, list-item]

    assert isinstance(aten_op, (list, tuple)), f"{type(aten_op)=}"
    for op in aten_op:
        assert isinstance(op, torch._ops.OpOverload)
        propagate_rules[op] = handler
    return handler


def register_propagate_rule(
    aten_op: torch._ops.OpOverload | Sequence[torch._ops.OpOverload],
) -> Callable[[_HandlerType], _HandlerType]:
    return functools.partial(_register_propagate_rule, aten_op)


def propagate_scale_by(nodes_with_chunking_meta: Sequence[Node]) -> None:
    """
    The input is a list of nodes that have chunking metadata.
    The nodes are already sorted in topological order.
    """
    for node in nodes_with_chunking_meta:
        arg_nodes = get_args_of_node_type(node)
        arg_metas = get_chunking_metas(arg_nodes)

        if all(arg_meta is None for arg_meta in arg_metas):
            # should be graph input of the chunking subgraph
            continue

        if log.isEnabledFor(logging.DEBUG):
            print("Propagate scale_by:")
            format_node_with_chunking_meta(node, True)

        assert all(arg_meta is not None for arg_meta in arg_metas), node.format_node()

        # None of the input has scale_by set
        if all(arg_meta.scale_by is None for arg_meta in arg_metas):  # type: ignore[union-attr]
            continue

        target = node.target
        if (
            not isinstance(target, torch._ops.OpOverload)
            or target not in propagate_rules
        ):
            raise CantChunk(
                f"Missing scale_by propagation rule for target {target}: {node.format_node()}"
            )

        if not propagate_rules[target](node):
            raise CantChunk(
                f"scale_by propagate rule for {target} fail: {node.format_node()}"
            )


@register_propagate_rule(
    [
        aten.div.Tensor,
    ]
)
def propagate_div(div_node: Node) -> bool:
    lhs_node, rhs_node = div_node.args[:2]
    assert isinstance(lhs_node, Node)
    lhs_scale_by = get_scale_by_from_node(lhs_node)

    # When gradient accumulation is enabled, rhs_node can be a constant
    # representing the gradient accumulation steps
    rhs_scale_by = (
        get_scale_by_from_node(rhs_node) if isinstance(rhs_node, Node) else None
    )
    if lhs_scale_by and rhs_scale_by is None:
        update_chunking_meta(div_node, scale_by=lhs_scale_by)
        return True
    return False


@register_propagate_rule(
    [
        aten.where.self,
    ]
)
def propagate_where(where_node: Node) -> bool:
    cond_node, true_node, false_node = where_node.args
    assert isinstance(cond_node, Node)
    assert isinstance(true_node, Node)
    assert isinstance(false_node, Node)
    cond_meta, true_meta, false_meta = get_chunking_metas(
        [cond_node, true_node, false_node]
    )
    out_meta = get_chunking_meta(where_node)

    assert true_meta is not None
    assert false_meta is not None
    if true_meta.scale_by and not false_meta.scale_by:
        # the false_node must be all zero
        if false_node.target != aten.full.default:
            return False
        if false_node.args[1] != 0.0:
            return False
        assert out_meta is not None
        out_meta.scale_by = true_meta.scale_by
        return True
    return False


@register_propagate_rule(
    [
        aten.exp.default,
        aten.log.default,
        aten.tanh.default,
        aten.eq.Tensor,
    ]
)
def propagate_requires_no_scaling(out_node: Node) -> bool:
    """
    For nonlinear ops (exp, log, tanh) scale_by cannot be propagated
    through since f(S*x) != S*f(x). For boolean-output ops (eq) scale_by
    is meaningless. These ops only appear in the chunking subgraph when
    scale_by is None (e.g. the final gradient is 1).
    """
    args_node = get_args_of_node_type(out_node)
    args_meta = get_chunking_metas(args_node)
    out_meta = get_chunking_meta(out_node)

    scale_by = get_scale_by_from_metas(*args_meta)  # type: ignore[arg-type]
    assert scale_by is None, (
        f"Nonlinear op {out_node.target} requires scale_by=None, got {scale_by}"
    )
    assert out_meta is not None
    out_meta.scale_by = None
    return True


@register_propagate_rule(
    [
        aten.mul.Tensor,
        prims.convert_element_type.default,
        aten.neg.default,
        aten.sum.dim_IntList,
        aten.sum.default,  # sum to scalar
        aten.amax.default,
        aten.mm.default,
        aten.permute.default,
        aten.expand.default,
        aten.squeeze.dim,
        aten.unsqueeze.default,
        aten.gather.default,
        aten.scatter_add.default,
        aten.view.default,
    ]
)
def propagate_general_copy(out_node: Node) -> bool:
    """
    A rule that holds for multiple ops: the scale_by of the output is
    set to the only scale_by of input nodes or None if no input has scale_by
    set.
    """
    args_node = get_args_of_node_type(out_node)
    args_meta = get_chunking_metas(args_node)
    out_meta = get_chunking_meta(out_node)

    scale_by = get_scale_by_from_metas(*args_meta)  # type: ignore[arg-type]
    assert out_meta is not None
    out_meta.scale_by = scale_by
    return True


@register_propagate_rule(aten.scatter.value)
def propagate_scatter_value(out_node: Node) -> bool:
    # The backward of scatter.value always has value=0 (gradient of a constant),
    # so S * scatter(x, idx, 0) = scatter(S*x, idx, 0) holds.
    value = out_node.args[3]
    if value != 0:
        return False
    return propagate_general_copy(out_node)


@register_propagate_rule(
    [
        aten.add.Tensor,
        aten.sub.Tensor,
    ]
)
def propagate_add_sub(out_node: Node) -> bool:
    """
    The scale_by node of the two arguments must be the same.
    """
    lhs_node, rhs_node = get_args_of_node_type(out_node)
    assert isinstance(lhs_node, Node)
    assert isinstance(rhs_node, Node)
    lhs_meta, rhs_meta = get_chunking_metas([lhs_node, rhs_node])
    assert lhs_meta is not None
    assert rhs_meta is not None
    if lhs_meta.scale_by is rhs_meta.scale_by:
        update_chunking_meta(out_node, scale_by=lhs_meta.scale_by)
        return True
    return False


@register_propagate_rule(
    [
        prims.fma.default,
    ]
)
def propagate_fma(out_node: Node) -> bool:
    mul_lhs, mul_rhs, add_rhs = out_node.args[:3]
    assert isinstance(mul_lhs, Node)
    assert isinstance(mul_rhs, Node)
    assert isinstance(add_rhs, Node)
    mul_lhs_meta, mul_rhs_meta, add_rhs_meta = get_chunking_metas(
        [mul_lhs, mul_rhs, add_rhs]
    )
    assert mul_lhs_meta is not None
    assert mul_rhs_meta is not None
    add_lhs_scale_by = get_scale_by_from_metas(mul_lhs_meta, mul_rhs_meta)
    assert add_rhs_meta is not None
    if add_lhs_scale_by is add_rhs_meta.scale_by:
        update_chunking_meta(out_node, scale_by=add_lhs_scale_by)
        return True
    return False



"""
This module implements TorchDynamo's backend registry system for managing compiler backends.

The registry provides a centralized way to register, discover and manage different compiler
backends that can be used with torch.compile(). It handles:

- Backend registration and discovery through decorators and entry points
- Lazy loading of backend implementations
- Lookup and validation of backend names
- Categorization of backends using tags (debug, experimental, etc.)

Key components:
- CompilerFn: Type for backend compiler functions that transform FX graphs
- _BACKENDS: Registry mapping backend names to entry points
- _COMPILER_FNS: Registry mapping backend names to loaded compiler functions

Example usage:
    @register_backend
    def my_compiler(fx_graph, example_inputs):
        # Transform FX graph into optimized implementation
        return compiled_fn

    # Use registered backend
    torch.compile(model, backend="my_compiler")

The registry also supports discovering backends through setuptools entry points
in the "torch_dynamo_backends" group. Example:
```
setup.py
---
from setuptools import setup

setup(
    name='my_torch_backend',
    version='0.1',
    packages=['my_torch_backend'],
    entry_points={
        'torch_dynamo_backends': [
            # name = path to entry point of backend implementation
            'my_compiler = my_torch_backend.compiler:my_compiler_function',
        ],
    },
)
```
```
my_torch_backend/compiler.py
---
def my_compiler_function(fx_graph, example_inputs):
    # Transform FX graph into optimized implementation
    return compiled_fn
```
Using `my_compiler` backend:
```
import torch

model = ...  # Your PyTorch model
optimized_model = torch.compile(model, backend="my_compiler")
```
"""

import functools
import logging
from collections.abc import Callable, Sequence
from importlib.metadata import EntryPoint
from typing import Any, Protocol

import torch
from torch import fx


log = logging.getLogger(__name__)


class CompiledFn(Protocol):
    def __call__(self, *args: torch.Tensor) -> tuple[torch.Tensor, ...]: ...


CompilerFn = Callable[[fx.GraphModule, list[torch.Tensor]], CompiledFn]

_BACKENDS: dict[str, EntryPoint | None] = {}
_COMPILER_FNS: dict[str, CompilerFn] = {}


def register_backend(
    compiler_fn: CompilerFn | None = None,
    name: str | None = None,
    tags: Sequence[str] = (),
) -> Callable[..., Any]:
    """
    Decorator to add a given compiler to the registry to allow calling
    `torch.compile` with string shorthand.  Note: for projects not
    imported by default, it might be easier to pass a function directly
    as a backend and not use a string.

    Args:
        compiler_fn: Callable taking a FX graph and fake tensor inputs
        name: Optional name, defaults to `compiler_fn.__name__`
        tags: Optional set of string tags to categorize backend with
    """
    if compiler_fn is None:
        # @register_backend(name="") syntax
        return functools.partial(register_backend, name=name, tags=tags)  # type: ignore[return-value]
    assert callable(compiler_fn)
    name = name or compiler_fn.__name__
    assert name not in _COMPILER_FNS, f"duplicate name: {name}"
    if compiler_fn not in _BACKENDS:
        _BACKENDS[name] = None
    _COMPILER_FNS[name] = compiler_fn
    compiler_fn._tags = tuple(tags)  # type: ignore[attr-defined]
    return compiler_fn


register_debug_backend = functools.partial(register_backend, tags=("debug",))
register_experimental_backend = functools.partial(
    register_backend, tags=("experimental",)
)


def lookup_backend(compiler_fn: str | CompilerFn) -> CompilerFn:
    """Expand backend strings to functions"""
    if isinstance(compiler_fn, str):
        if compiler_fn not in _BACKENDS:
            _lazy_import()
        if compiler_fn not in _BACKENDS:
            from ..exc import InvalidBackend

            raise InvalidBackend(name=compiler_fn)

        if compiler_fn not in _COMPILER_FNS:
            entry_point = _BACKENDS[compiler_fn]
            if entry_point is not None:
                register_backend(compiler_fn=entry_point.load(), name=compiler_fn)
        compiler_fn = _COMPILER_FNS[compiler_fn]
    return compiler_fn


# NOTE: can't type this due to public api mismatch; follow up with dev team
def list_backends(exclude_tags=("debug", "experimental")) -> list[str]:  # type: ignore[no-untyped-def]
    """
    Return valid strings that can be passed to:

        torch.compile(..., backend="name")
    """
    _lazy_import()
    exclude_tags_set = set(exclude_tags or ())

    backends = [
        name
        for name in _BACKENDS
        if name not in _COMPILER_FNS
        or not exclude_tags_set.intersection(_COMPILER_FNS[name]._tags)  # type: ignore[attr-defined]
    ]
    return sorted(backends)


@functools.cache
def _lazy_import() -> None:
    from .. import backends
    from ..utils import import_submodule

    import_submodule(backends)

    from ..repro.after_dynamo import dynamo_minifier_backend

    assert dynamo_minifier_backend is not None

    _discover_entrypoint_backends()


@functools.cache
def _discover_entrypoint_backends() -> None:
    # importing here so it will pick up the mocked version in test_backends.py
    from importlib.metadata import entry_points

    group_name = "torch_dynamo_backends"
    eps = entry_points(group=group_name)
    # pyrefly: ignore [bad-index]
    eps_dict = {name: eps[name] for name in eps.names}
    for backend_name in eps_dict:
        _BACKENDS[backend_name] = eps_dict[backend_name]


def _is_registered_backend(compiler_fn: CompilerFn) -> bool:
    """
    Check if the given compiler function is a registered backend.
    Custom backends (user-provided callables not in the registry) return False.
    """
    # Ensure backends are loaded
    _lazy_import()

    # Check if it's directly a registered backend function
    if compiler_fn in _COMPILER_FNS.values():
        return True

    # Check for _TorchCompileInductorWrapper or _TorchCompileWrapper
    # These have a compiler_name attribute that identifies the backend
    if hasattr(compiler_fn, "compiler_name"):
        compiler_name = compiler_fn.compiler_name
        if compiler_name in _BACKENDS or compiler_name in _COMPILER_FNS:
            return True

    # Check if the wrapper has a compiler_fn attribute (e.g., _TorchCompileWrapper)
    if hasattr(compiler_fn, "compiler_fn"):
        return compiler_fn.compiler_fn in _COMPILER_FNS.values()

    return False



# mypy: allow-untyped-defs
import torch._C
from torch.distributed.tensor.debug._comm_mode import CommDebugMode
from torch.distributed.tensor.debug._visualize_sharding import visualize_sharding


__all__ = ["CommDebugMode", "visualize_sharding"]


def _get_python_sharding_prop_cache_info():
    """
    Get the cache info for the Python sharding propagation cache, used for debugging purpose only.
    This would return a named tuple showing hits, misses, maxsize and cursize of the sharding
    propagator cache. Note that directly calling into the sharding propagator does not share cache
    state with the DTensor dispatch fast path!
    """
    from torch.distributed.tensor._api import DTensor

    return (
        DTensor._op_dispatcher.sharding_propagator.propagate_op_sharding.cache_info()  # type:ignore[attr-defined]
    )


def _get_fast_path_sharding_prop_cache_stats():
    """
    Get a tuple (hits, misses) for the fast path sharding propagation cache, used for debugging
    only.
    """
    return torch._C._get_DTensor_sharding_propagator_cache_stats()


def _clear_python_sharding_prop_cache():
    """
    Clears the cache for the Python sharding propagation cache, used for debugging purpose only.
    """
    from torch.distributed.tensor._api import DTensor

    return (
        DTensor._op_dispatcher.sharding_propagator.propagate_op_sharding.cache_clear()  # type:ignore[attr-defined]
    )


def _clear_fast_path_sharding_prop_cache():
    """
    Clears the cache for the fast path sharding propagation cache, used for debugging purpose only.
    """
    torch._C._clear_DTensor_sharding_propagator_cache()


def _clear_sharding_prop_cache():
    """
    Clears both the Python and fast path sharding propagation caches, used for debugging purpose only.
    This is the recommended way to clear all sharding propagation caches.
    """
    _clear_python_sharding_prop_cache()
    _clear_fast_path_sharding_prop_cache()


def _reinit_dispatch_logger():
    """
    Resets the cached DTensor dispatch logger state so that the next DTensor
    dispatch re-checks whether debug logging is enabled. Call this after
    changing the log level on the ``torch.distributed.tensor._dispatch`` logger.
    """
    torch._C._reinit_DTensor_dispatch_logger()


# Set namespace for exposed private names
CommDebugMode.__module__ = "torch.distributed.tensor.debug"
visualize_sharding.__module__ = "torch.distributed.tensor.debug"



# Copyright (c) Meta Platforms, Inc. and affiliates
# Owner(s): ["oncall: distributed"]

import numpy as np

from torch.distributed.device_mesh import DeviceMesh, init_device_mesh
from torch.testing._internal.common_utils import run_tests
from torch.testing._internal.distributed._tensor.common_dtensor import (
    DTensorTestBase,
    with_comms,
)


class DTensorTestBaseUtilCPUTest(DTensorTestBase):
    """
    This class tests if the basic functionalities of DTensorTestBase are
    working as expected on CPU, regardless of the presence of CUDA devices.
    """

    @property
    def backend(self):
        return "gloo"

    @property
    def device_type(self) -> str:
        return "cpu"

    @property
    def world_size(self):
        return np.prod(list(self.mesh_dim_sizes.values())).item()

    @property
    def mesh_dim_sizes(self) -> dict[str, int]:
        """Mapping from mesh dimension names to sizes."""
        return {"data": 2, "fsdp": 3, "tensor": 5}

    def build_device_mesh(self) -> DeviceMesh:
        return init_device_mesh(
            self.device_type,
            mesh_shape=tuple(self.mesh_dim_sizes.values()),
            mesh_dim_names=tuple(self.mesh_dim_sizes.keys()),
        )

    @with_comms
    def test_dtensor_testbase_destroy_pg(self):
        # This tests destroy_pg() correctly finishes.
        device_mesh = self.build_device_mesh()  # noqa: F841


if __name__ == "__main__":
    run_tests()



#!/usr/bin/env python3
import argparse
import os
import subprocess
from pathlib import Path

from common import (
    get_testcases,
    is_failure,
    is_passing_skipped_test,
    is_unexpected_success,
    key,
    open_test_results,
)
from download_reports import download_reports


"""
Usage: update_failures.py /path/to/dynamo_test_failures.py /path/to/test commit_sha

Best-effort updates the xfail and skip files under test directory
by parsing test reports.

You'll need to provide the commit_sha for the latest commit on a PR
from which we will pull CI test results.

Instructions:
- On your PR, add the "keep-going" label to ensure that all the tests are
  failing (as opposed to CI stopping on the first failure). You may need to
  restart your test jobs by force-pushing to your branch for CI to pick
  up the "keep-going" label.
- Wait for all the tests to finish running.
- Find the full SHA of your commit and run this command.

This script requires the `gh` cli. You'll need to install it and then
authenticate with it via `gh auth login` before using this script.
https://docs.github.com/en/github-cli/github-cli/quickstart
"""


def patch_file(
    filename, test_dir, unexpected_successes, new_xfails, new_skips, unexpected_skips
):
    failures_directory = os.path.join(test_dir, "dynamo_expected_failures")
    skips_directory = os.path.join(test_dir, "dynamo_skips")

    dynamo_expected_failures = set(os.listdir(failures_directory))
    dynamo_skips = set(os.listdir(skips_directory))

    # These are hand written skips
    extra_dynamo_skips = set()
    with open(filename) as f:
        start = False
        for text in f.readlines():
            text = text.strip()
            if start:
                if text == "}":
                    break
                extra_dynamo_skips.add(text.strip(',"'))
            else:
                if text == "extra_dynamo_skips = {":
                    start = True

    def format(testcase):
        classname = testcase.attrib["classname"]
        name = testcase.attrib["name"]
        return f"{classname}.{name}"

    formatted_unexpected_successes = {
        f"{format(test)}" for test in unexpected_successes.values()
    }
    formatted_unexpected_skips = {
        f"{format(test)}" for test in unexpected_skips.values()
    }
    formatted_new_xfails = [f"{format(test)}" for test in new_xfails.values()]
    formatted_new_skips = [f"{format(test)}" for test in new_skips.values()]

    def remove_file(path, name):
        file = os.path.join(path, name)
        cmd = ["git", "rm", file]
        subprocess.run(cmd)

    def add_file(path, name):
        file = os.path.join(path, name)
        with open(file, "w") as fp:
            pass
        cmd = ["git", "add", file]
        subprocess.run(cmd)

    covered_unexpected_successes = set()

    # dynamo_expected_failures
    for test in dynamo_expected_failures:
        if test in formatted_unexpected_successes:
            covered_unexpected_successes.add(test)
            remove_file(failures_directory, test)
    for test in formatted_new_xfails:
        add_file(failures_directory, test)

    leftover_unexpected_successes = (
        formatted_unexpected_successes - covered_unexpected_successes
    )
    if len(leftover_unexpected_successes) > 0:
        print(
            "WARNING: we were unable to remove these "
            f"{len(leftover_unexpected_successes)} expectedFailures:"
        )
        for stuff in leftover_unexpected_successes:
            print(stuff)

    # dynamo_skips
    for test in dynamo_skips:
        if test in formatted_unexpected_skips:
            remove_file(skips_directory, test)
    for test in extra_dynamo_skips:
        if test in formatted_unexpected_skips:
            print(
                f"WARNING: {test} in dynamo_test_failures.py needs to be removed manually"
            )
    for test in formatted_new_skips:
        add_file(skips_directory, test)


def get_intersection_and_outside(a_dict, b_dict):
    a = set(a_dict.keys())
    b = set(b_dict.keys())
    intersection = a.intersection(b)
    outside = (a.union(b)) - intersection

    def build_dict(keys):
        result = {}
        for k in keys:
            if k in a_dict:
                result[k] = a_dict[k]
            else:
                result[k] = b_dict[k]
        return result

    return build_dict(intersection), build_dict(outside)


def update(filename, test_dir, py38_dir, py311_dir, also_remove_skips):
    def read_test_results(directory):
        xmls = open_test_results(directory)
        testcases = get_testcases(xmls)
        unexpected_successes = {
            key(test): test for test in testcases if is_unexpected_success(test)
        }
        failures = {key(test): test for test in testcases if is_failure(test)}
        passing_skipped_tests = {
            key(test): test for test in testcases if is_passing_skipped_test(test)
        }
        return unexpected_successes, failures, passing_skipped_tests

    (
        py38_unexpected_successes,
        py38_failures,
        py38_passing_skipped_tests,
    ) = read_test_results(py38_dir)
    (
        py311_unexpected_successes,
        py311_failures,
        py311_passing_skipped_tests,
    ) = read_test_results(py311_dir)

    unexpected_successes = {**py38_unexpected_successes, **py311_unexpected_successes}
    _, skips = get_intersection_and_outside(
        py38_unexpected_successes, py311_unexpected_successes
    )
    xfails, more_skips = get_intersection_and_outside(py38_failures, py311_failures)
    if also_remove_skips:
        unexpected_skips, _ = get_intersection_and_outside(
            py38_passing_skipped_tests, py311_passing_skipped_tests
        )
    else:
        unexpected_skips = {}
    all_skips = {**skips, **more_skips}
    print(
        f"Discovered {len(unexpected_successes)} new unexpected successes, "
        f"{len(xfails)} new xfails, {len(all_skips)} new skips, {len(unexpected_skips)} new unexpected skips"
    )
    return patch_file(
        filename, test_dir, unexpected_successes, xfails, all_skips, unexpected_skips
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="update_dynamo_test_failures",
        description="Read from logs and update the dynamo_test_failures file",
    )
    # dynamo_test_failures path
    parser.add_argument(
        "filename",
        nargs="?",
        default=str(
            Path(__file__).absolute().parents[2]
            / "torch/testing/_internal/dynamo_test_failures.py"
        ),
        help="Optional path to dynamo_test_failures.py",
    )
    # test path
    parser.add_argument(
        "test_dir",
        nargs="?",
        default=str(Path(__file__).absolute().parents[2] / "test"),
        help="Optional path to test folder",
    )
    parser.add_argument(
        "commit",
        help=(
            "The commit sha for the latest commit on a PR from which we will "
            "pull CI test results, e.g. 7e5f597aeeba30c390c05f7d316829b3798064a5"
        ),
    )
    parser.add_argument(
        "--also-remove-skips",
        help="Also attempt to remove skips. WARNING: does not guard against test flakiness",
        action="store_true",
    )
    args = parser.parse_args()
    assert Path(args.filename).exists(), args.filename
    assert Path(args.test_dir).exists(), args.test_dir
    dynamo39, dynamo313 = download_reports(args.commit, ("dynamo39", "dynamo313"))
    update(args.filename, args.test_dir, dynamo39, dynamo313, args.also_remove_skips)



# Copyright (c) Meta Platforms, Inc. and affiliates
# Owner(s): ["oncall: distributed"]

from collections.abc import Callable, Sequence
from typing import Any
from unittest import skip

import torch
import torch.utils._pytree as pytree
from torch import Tensor
from torch.distributed.tensor import (
    DeviceMesh,
    distribute_tensor,
    DTensor,
    init_device_mesh,
    Partial,
    Placement,
    Replicate,
    Shard,
)
from torch.distributed.tensor._ops._math_ops import _NormPartial
from torch.distributed.tensor.debug import CommDebugMode
from torch.distributed.tensor.placement_types import _StridedShard
from torch.testing._internal.common_utils import (
    instantiate_parametrized_tests,
    parametrize,
    run_tests,
    TestCase,
)
from torch.testing._internal.distributed._tensor.common_dtensor import (
    create_local_tensor_test_class,
    DTensorOpTestBase,
    LocalDTensorOpTestBase,
    map_local_for_rank,
    skip_unless_torch_gpu,
    with_comms,
)


def no_op():
    return None


def deepcopy_convert_to_dtensor(
    val: Any,
    device_mesh: DeviceMesh,
    placements: Sequence[Placement],
) -> Any:
    """
    Recursively convert (over Sequence and Dict types) Tensors into DTensors.

    :param device_mesh: the DeviceMesh to use.
    :param placements: the Placement list to use.
    :return: the transformed structure.
    """

    def f(x):
        if isinstance(x, Tensor) and not isinstance(x, DTensor):
            return distribute_tensor(
                x,
                device_mesh=device_mesh,
                placements=placements,
            )
        return x

    return pytree.tree_map(f, [val])[0]


def deepcopy_convert_from_dtensor(val: Any) -> Any:
    """
    Recursive convert any DTensor to local Tensor.

    :param val: the structure to coerce.
    :return: the coerced structure.
    """

    def f(x):
        if isinstance(x, DTensor):
            return x.full_tensor()
        return x

    return pytree.tree_map(f, [val])[0]


class DistElementwiseOpsTest(DTensorOpTestBase):
    def _compare_pairwise_ops(
        self,
        *,
        device_mesh: DeviceMesh,
        placements: Sequence[Placement],
        op: Callable,
        pre_op_fn: Callable | None = None,
        args: Sequence[Any] = (),
        kwargs: dict[str, Any] | None = None,
    ):
        if pre_op_fn is None:
            pre_op_fn = no_op

        if not kwargs:
            kwargs = {}

        dargs = deepcopy_convert_to_dtensor(
            args,
            device_mesh=device_mesh,
            placements=placements,
        )
        dkwargs = deepcopy_convert_to_dtensor(
            kwargs,
            device_mesh=device_mesh,
            placements=placements,
        )

        pre_op_fn()

        # run the reference first, in case the call is broken;
        # it's better to debug an incorrect call at this point.
        reference_result = op(*args, **kwargs)

        pre_op_fn()

        dist_result = op(*dargs, **dkwargs)

        collected_result = deepcopy_convert_from_dtensor(dist_result)

        self.assertEqualOnRank(reference_result, collected_result)

    # TODO: We need to add CPU tests for ops in the future.
    def _run_sharded_elementwise_ops(
        self,
        *,
        device_mesh: DeviceMesh,
        placements: Sequence[Placement],
        pre_op_fn: Callable | None = None,
        input_size: Sequence[int],
        op: Callable,
        **kwargs,
    ):
        if pre_op_fn is None:
            pre_op_fn = no_op

        input_tensor = torch.randn(
            *input_size,
            device=self.device_type,
            requires_grad=True,
        )

        self._compare_pairwise_ops(
            device_mesh=device_mesh,
            placements=placements,
            pre_op_fn=pre_op_fn,
            op=op,
            args=(input_tensor,),
            kwargs=kwargs,
        )

    @with_comms
    @skip("This test will sometimes hang and a timeout error will be thrown")
    def test_partial_add(self):
        device_mesh = self.build_device_mesh()
        d_1 = DTensor.from_local(torch.rand(2, 2), device_mesh, [Partial()])
        d_2 = DTensor.from_local(torch.rand(2, 2), device_mesh, [Partial()])
        d_3 = d_1 + d_2
        self.assertTrue(d_3._spec.placements[0].is_partial())

    @with_comms
    def test_partial_replicate_add(self):
        device_mesh = self.build_device_mesh()
        comm_mode = CommDebugMode()

        for reduce_op in ("sum", "avg"):
            d_1 = DTensor.from_local(
                torch.rand(2, 2),
                device_mesh,
                [Partial(reduce_op=reduce_op)],
            )
            d_2 = DTensor.from_local(
                torch.rand(2, 1),
                device_mesh,
                [Replicate()],
                run_check=True,
            )

            with comm_mode:
                d_3 = d_1 + d_2

            self.assertEqual(comm_mode.get_total_counts(), 0)
            self.assertEqual(d_3.placements, (Partial(reduce_op=reduce_op),))
            self.assertEqual(d_3.full_tensor(), d_1.full_tensor() + d_2.full_tensor())

    @with_comms
    def test_activations(self):
        device_mesh = self.build_device_mesh()
        self._run_sharded_elementwise_ops(
            device_mesh=device_mesh,
            placements=[Shard(0)],
            input_size=(8, 5),
            op=torch.nn.functional.gelu,
        )
        self._run_sharded_elementwise_ops(
            device_mesh=device_mesh,
            placements=[Replicate()],
            input_size=(8, 5),
            op=torch.nn.functional.gelu,
        )
        self._run_sharded_elementwise_ops(
            device_mesh=device_mesh,
            placements=[Shard(1)],
            input_size=(3, 12),
            op=torch.nn.functional.relu,
        )
        self._run_sharded_elementwise_ops(
            device_mesh=device_mesh,
            placements=[Replicate()],
            input_size=(8, 5),
            op=torch.nn.functional.relu,
        )
        self._run_sharded_elementwise_ops(
            device_mesh=device_mesh,
            placements=[Shard(0)],
            input_size=(8, 5),
            op=torch.sigmoid,
        )
        self._run_sharded_elementwise_ops(
            device_mesh=device_mesh,
            placements=[Replicate()],
            input_size=(8, 5),
            op=torch.sigmoid,
        )

    @with_comms
    @skip(
        "testing RNG based ops is broken: https://github.com/pytorch/PiPPy/issues/494"
    )
    def test_dropout(self):
        device_mesh = self.build_device_mesh()

        def _reset_random_seed():
            torch.manual_seed(self.rank + 4)

        self._run_sharded_elementwise_ops(
            device_mesh=device_mesh,
            placements=[Shard(0)],
            input_size=(8, 5),
            op=torch.nn.functional.dropout,
            pre_op_fn=_reset_random_seed,
            p=0.4,
            training=False,
        )
        self._run_sharded_elementwise_ops(
            device_mesh=device_mesh,
            placements=[Shard(1)],
            input_size=(3, 14),
            op=torch.nn.functional.dropout,
            pre_op_fn=_reset_random_seed,
            p=0.5,
            training=True,
        )

    @with_comms
    @skip_unless_torch_gpu
    def test_dropout_backward(self):
        device_mesh = self.build_device_mesh()
        placements = [Shard(0)]

        input_size = (8, 5)

        grad_output = torch.rand(
            input_size,
            device=self.device_type,
            requires_grad=True,
        )
        mask = (
            torch.rand(
                input_size,
                device=self.device_type,
                requires_grad=False,
            )
            < 0.8
        )

        self._compare_pairwise_ops(
            device_mesh=device_mesh,
            placements=placements,
            op=torch.ops.aten.native_dropout_backward,
            kwargs=dict(
                grad_output=grad_output,
                mask=mask,
                scale=0.3,
            ),
        )

    @with_comms
    @skip_unless_torch_gpu
    def test_dropout_errors(self):
        device_mesh = self.build_device_mesh()
        with self.assertRaisesRegex(RuntimeError, "supported"):
            self._run_sharded_elementwise_ops(
                device_mesh=device_mesh,
                placements=[Partial("sum")],
                input_size=(8, 5),
                op=torch.nn.functional.dropout,
            )

    @with_comms
    def test_mul_out(self):
        device_mesh = self.build_device_mesh()
        torch.manual_seed(self.rank)
        shard_spec = [Shard(0)]
        input_size = (8, 4)
        input_tensor = torch.randn(*input_size, device=self.device_type)
        dtensor = DTensor.from_local(input_tensor, device_mesh, shard_spec)

        other_tensor = torch.randn(*input_size, device=self.device_type)
        other_dtensor = DTensor.from_local(other_tensor, device_mesh, shard_spec)

        output_tensor = torch.randn(*input_size, device=self.device_type)
        output_dtensor = DTensor.from_local(output_tensor, device_mesh, shard_spec)
        dt = torch.mul(dtensor, other_dtensor, out=output_dtensor)
        expected = torch.mul(input_tensor, other_tensor, out=output_tensor)
        self.assertEqual(input_tensor, dtensor.to_local())
        self.assertEqual(expected, dt.to_local())

    @with_comms
    def test_mul_out_partial(self):
        device_mesh = self.build_device_mesh()
        input_size = (8, 4)
        # P(sum) * R -> P(sum), with out= kwarg
        partial_tensor = DTensor.from_local(
            torch.ones(*input_size, device=self.device_type),
            device_mesh,
            [Partial("sum")],
        )
        replicate_tensor = DTensor.from_local(
            torch.full(input_size, 2.0, device=self.device_type),
            device_mesh,
            [Replicate()],
            run_check=False,
        )
        output_tensor = DTensor.from_local(
            torch.empty(*input_size, device=self.device_type),
            device_mesh,
            [Partial("sum")],
        )
        dt = torch.mul(partial_tensor, replicate_tensor, out=output_tensor)
        self.assertEqual(dt.placements, (Partial("sum"),))
        self.assertEqual(output_tensor.placements, (Partial("sum"),))

    @with_comms
    def test_mul_partial(self):
        # we only test the partial behavior for mul op as other placement
        # behaviors should be well tested in test_dtensor_ops.py
        device_mesh = self.build_device_mesh()
        comm_mode = CommDebugMode()
        # 1. simple test for partial * partial
        d_1 = DTensor.from_local(torch.ones(2, 2), device_mesh, [Partial()])
        d_2 = DTensor.from_local(torch.ones(2, 2), device_mesh, [Partial()])
        with comm_mode:
            d_3 = d_1 * d_2
        comm_counts = comm_mode.get_total_counts()
        self.assertEqual(comm_counts, 1)
        self.assertTrue(isinstance(d_3, DTensor))
        self.assertEqual(d_3.placements, (Partial(),))
        self.assertEqual(d_3.to_local(), torch.ones(2, 2) * (self.world_size))

        # 2. test the partial input DTensor * scalar/replicate input
        input = torch.full((8, 8), 1.0, device=self.device_type)

        # test for different types of other inputs
        other_inps = (
            2.0,  # scalar
            torch.tensor(2.0, device=self.device_type),  # scalar tensor
            torch.full((8, 8), 2.0, device=self.device_type),  # tensor
        )

        for partial_op in ["sum", "avg"]:
            expected_p_out = (
                input * self.world_size * 2.0 if partial_op == "sum" else input * 2.0
            )

            d_input = DTensor.from_local(input, device_mesh, [Partial(partial_op)])

            for other_inp in other_inps:
                if isinstance(other_inp, Tensor) and other_inp.numel() > 1:
                    d_other = distribute_tensor(other_inp, device_mesh, [Replicate()])
                else:
                    d_other = other_inp

                with comm_mode:
                    z = d_input * d_other

                comm_counts = comm_mode.get_total_counts()
                self.assertEqual(comm_counts, 0)
                self.assertTrue(isinstance(z, DTensor))
                self.assertEqual(z.placements, (Partial(partial_op),))
                self.assertEqual(z.full_tensor(), expected_p_out)

        # test other partial to assert the partial not getting propagated
        d_input = DTensor.from_local(input, device_mesh, [Partial("max")])
        d_other = distribute_tensor(torch.ones(8, 8), device_mesh, [Replicate()])

        z = d_input * d_other
        self.assertEqual(z.placements, (Replicate(),))
        self.assertEqual(z.to_local(), input)

    @with_comms
    def test_div_partial(self):
        # we only test the partial behavior for div op as other placement
        # behaviors should be well tested in test_dtensor_ops.py
        device_mesh = self.build_device_mesh()
        comm_mode = CommDebugMode()

        # 1. test the partial input DTensor / scalar/replicate input
        # This is mathematically sound: (a1 + a2) / b = a1/b + a2/b
        input = torch.full((8, 8), 2.0, device=self.device_type)

        # test for different types of other inputs
        other_inps = (
            2.0,  # scalar
            torch.tensor(2.0, device=self.device_type),  # scalar tensor
            torch.full((8, 8), 2.0, device=self.device_type),  # tensor
        )

        for partial_op in ["sum", "avg"]:
            expected_p_out = (
                input * self.world_size / 2.0 if partial_op == "sum" else input / 2.0
            )

            d_input = DTensor.from_local(input, device_mesh, [Partial(partial_op)])

            for other_inp in other_inps:
                if isinstance(other_inp, Tensor) and other_inp.numel() > 1:
                    d_other = distribute_tensor(other_inp, device_mesh, [Replicate()])
                else:
                    d_other = other_inp

                with comm_mode:
                    z = d_input / d_other

                comm_counts = comm_mode.get_total_counts()
                self.assertEqual(comm_counts, 0)
                self.assertTrue(isinstance(z, DTensor))
                self.assertEqual(z.placements, (Partial(partial_op),))
                self.assertEqual(z.full_tensor(), expected_p_out)

        # test other partial to assert the partial not getting propagated
        d_input = DTensor.from_local(input, device_mesh, [Partial("max")])
        d_other = distribute_tensor(
            torch.full((8, 8), 2.0, device=self.device_type), device_mesh, [Replicate()]
        )

        z = d_input / d_other
        self.assertEqual(z.placements, (Replicate(),))
        self.assertEqual(z.to_local(), input / 2.0)

    @with_comms
    def test_masked_fill_scalar(self):
        """Test masked_fill_ with scalar value."""
        device_mesh = self.build_device_mesh()

        # Test with deterministic values to avoid random seed issues in threaded tests
        # Test with Shard(0) placement
        input_tensor = torch.arange(
            40, dtype=torch.float32, device=self.device_type
        ).reshape(8, 5)
        mask = input_tensor > 20
        fill_value = -999.0

        # Create DTensor
        dt_input = distribute_tensor(input_tensor.clone(), device_mesh, [Shard(0)])
        dt_mask = distribute_tensor(mask, device_mesh, [Shard(0)])

        # Perform in-place masked_fill
        input_tensor.masked_fill_(mask, fill_value)
        dt_input.masked_fill_(dt_mask, fill_value)

        # Compare results
        self.assertEqual(input_tensor, dt_input.full_tensor())

        # Test with Replicate placement
        input_tensor2 = (
            torch.arange(40, dtype=torch.float32, device=self.device_type).reshape(8, 5)
            - 20
        )
        mask2 = input_tensor2 < 0
        fill_value2 = 42.0

        dt_input2 = distribute_tensor(input_tensor2.clone(), device_mesh, [Replicate()])
        dt_mask2 = distribute_tensor(mask2, device_mesh, [Replicate()])

        input_tensor2.masked_fill_(mask2, fill_value2)
        dt_input2.masked_fill_(dt_mask2, fill_value2)

        self.assertEqual(input_tensor2, dt_input2.full_tensor())

        # Test with Shard(1) placement
        input_tensor3 = torch.arange(
            48, dtype=torch.float32, device=self.device_type
        ).reshape(4, 12)
        mask3 = input_tensor3 % 2 == 0  # even numbers
        fill_value3 = 0.0

        dt_input3 = distribute_tensor(input_tensor3.clone(), device_mesh, [Shard(1)])
        dt_mask3 = distribute_tensor(mask3, device_mesh, [Shard(1)])

        input_tensor3.masked_fill_(mask3, fill_value3)
        dt_input3.masked_fill_(dt_mask3, fill_value3)

        self.assertEqual(input_tensor3, dt_input3.full_tensor())

    @with_comms
    def test_inplace_op_partial_to_replicate(self):
        # test that in-place operations that require redistribution raise an error
        # to preserve aliasing semantics (issue #163374)
        device_mesh = self.build_device_mesh()

        input_tensor = torch.tensor(64.0, device=self.device_type)
        partial_dt = DTensor.from_local(
            input_tensor, device_mesh, placements=(Partial(),)
        )

        self.assertTrue(partial_dt.placements[0].is_partial())

        # Inplace ops that require placement changes (Partial -> Replicate) should error
        with self.assertRaisesRegex(
            RuntimeError,
            "in-place operations that require placement changes are not supported",
        ):
            partial_dt.clamp_(max=10)

    @with_comms
    def test_mul_div_scalar_partial(self):
        aten = torch.ops.aten
        mesh = self.build_device_mesh()

        # regular partial *,/ scalar
        local_tensor = map_local_for_rank(self.rank, lambda rank: torch.tensor([rank]))

        dt = DTensor.from_local(
            local_tensor, device_mesh=mesh, placements=[Partial("sum")]
        )

        res = aten.mul.Scalar(dt, 2)
        self.assertEqual(
            res.to_local(),
            map_local_for_rank(self.rank, lambda rank: torch.tensor([rank * 2])),
        )

        self.assertTrue(res._spec.placements[0].is_partial())
        res = res.redistribute(dt.device_mesh, placements=[Replicate()])
        expected = sum(i for i in range(self.world_size)) * 2
        self.assertEqual(res.full_tensor(), expected)

        res = aten.div.Scalar(dt, 2)
        self.assertEqual(
            res.to_local(),
            map_local_for_rank(self.rank, lambda rank: torch.tensor([rank / 2])),
        )

        self.assertTrue(res._spec.placements[0].is_partial())
        res = res.redistribute(dt.device_mesh, placements=[Replicate()])
        expected = sum(i for i in range(self.world_size)) / 2
        self.assertEqual(res.full_tensor(), expected)

    @with_comms
    def test_mul_div_scalar_norm_partial(self):
        mesh = self.build_device_mesh()
        aten = torch.ops.aten
        local_tensor = torch.tensor([1.0, 1.0, 7.0, 7.0])
        dt = distribute_tensor(local_tensor, mesh, [Shard(0)])

        norm = dt.norm()
        self.assertTrue(isinstance(norm._spec.placements[0], _NormPartial))

        self.assertEqual(aten.mul.Scalar(norm, 2).full_tensor(), 20)
        self.assertEqual(aten.div.Scalar(norm, 2).full_tensor(), 5)
        self.assertEqual(aten.mul.Scalar(norm, -2).full_tensor(), -20)
        self.assertEqual(aten.div.Scalar(norm, -2).full_tensor(), -5)

    @with_comms
    def test_add_sub_scalar_partial(self):
        mesh = self.build_device_mesh()

        rank = self.rank

        # regular partial + scalar -> replicate
        local_tensor = map_local_for_rank(rank, lambda rank: torch.tensor([rank]))

        dt = DTensor.from_local(
            local_tensor, device_mesh=mesh, placements=[Partial("sum")]
        )

        res = dt + 1
        expected = sum(i for i in range(self.world_size)) + 1
        self.assertEqual(res.full_tensor(), expected)
        self.assertTrue(res._spec.placements[0].is_replicate())

        # regular partial - scalar -> replicate
        local_tensor = map_local_for_rank(rank, lambda rank: torch.tensor([rank]))

        dt = DTensor.from_local(
            local_tensor, device_mesh=mesh, placements=[Partial("sum")]
        )

        res = dt - 1
        expected = sum(i for i in range(self.world_size)) - 1
        self.assertEqual(res.full_tensor(), expected)
        self.assertTrue(res._spec.placements[0].is_replicate())

        res = 7 - dt
        expected = 7 - sum(i for i in range(self.world_size))
        self.assertEqual(res.full_tensor(), expected)
        self.assertTrue(res._spec.placements[0].is_replicate())

        # regular partial + regular partial -> partial
        res = dt + dt
        self.assertEqual(res.to_local(), rank + rank)
        self.assertTrue(res._spec.placements[0].is_partial())
        res = res.redistribute(dt.device_mesh, placements=[Replicate()])
        expected = sum(i for i in range(self.world_size)) * 2
        self.assertEqual(res.full_tensor(), expected)

        # regular partial - regular partial -> partial
        res = dt - dt
        self.assertEqual(res.to_local(), rank - rank)
        self.assertTrue(res._spec.placements[0].is_partial())
        res = res.redistribute(dt.device_mesh, placements=[Replicate()])
        self.assertEqual(res.full_tensor(), 0)

    @with_comms
    def test_add_sub_scalar_norm_partial(self):
        mesh = self.build_device_mesh()

        # norm partial + scalar
        local_tensor = torch.tensor([1.0, 1.0, 7.0, 7.0])
        dt = distribute_tensor(local_tensor, mesh, [Shard(0)])

        norm = dt.norm()
        self.assertTrue(isinstance(norm._spec.placements[0], _NormPartial))
        norm = norm + 1

        self.assertEqual(norm.full_tensor(), 11)
        self.assertTrue(norm._spec.placements[0].is_replicate())

        dt = distribute_tensor(local_tensor, mesh, [Shard(0)])

        norm = dt.norm()
        self.assertTrue(isinstance(norm._spec.placements[0], _NormPartial))
        norm = norm - 1

        self.assertEqual(norm.full_tensor(), 9)
        self.assertTrue(norm._spec.placements[0].is_replicate())

    @with_comms
    @parametrize("op,reduce_op", [(torch.maximum, "max"), (torch.minimum, "min")])
    def test_partial_propagation(self, op, reduce_op):
        # Test that torch.maximum/minimum preserves Partial("max"/"min") placements
        # since max(max(a), max(b)) == max(a, b) and min(min(a), min(b)) == min(a, b)
        device_mesh = self.build_device_mesh()
        comm_mode = CommDebugMode()

        input1 = torch.rand(8, 8) * self.rank
        input2 = torch.rand(8, 8) * (self.world_size - self.rank)

        d_input1 = DTensor.from_local(input1, device_mesh, [Partial(reduce_op)])
        d_input2 = DTensor.from_local(input2, device_mesh, [Partial(reduce_op)])

        with comm_mode:
            result = op(d_input1, d_input2)

        # Should not require any communication
        self.assertEqual(comm_mode.get_total_counts(), 0)
        # Result should still be Partial with the same reduce_op
        self.assertEqual(result.placements, (Partial(reduce_op),))

    @with_comms
    def test_neg_partial(self):
        # test that neg preserves Partial placement without communication
        # math: -(A1 + A2) = -A1 + -A2
        device_mesh = self.build_device_mesh()
        comm_mode = CommDebugMode()

        input = torch.full((8, 8), 2.0, device=self.device_type)

        for partial_op in ["sum", "avg"]:
            expected_full = (
                torch.full((8, 8), -2.0 * self.world_size, device=self.device_type)
                if partial_op == "sum"
                else torch.full((8, 8), -2.0, device=self.device_type)
            )

            d_input = DTensor.from_local(input, device_mesh, [Partial(partial_op)])

            with comm_mode:
                z = torch.neg(d_input)

            comm_counts = comm_mode.get_total_counts()
            self.assertEqual(comm_counts, 0)
            self.assertTrue(isinstance(z, DTensor))
            self.assertEqual(z.placements, (Partial(partial_op),))
            self.assertEqual(z.full_tensor(), expected_full)

        # test non-sum/avg partial: neg is monotone decreasing, so
        # -max(A1, A2) = min(-A1, -A2), i.e. Partial("max") -> Partial("min")
        d_input = DTensor.from_local(input, device_mesh, [Partial("max")])

        with comm_mode:
            z = torch.neg(d_input)

        comm_counts = comm_mode.get_total_counts()
        self.assertEqual(comm_counts, 0)
        self.assertEqual(z.placements, (Partial("min"),))

        d_input = DTensor.from_local(input, device_mesh, [Partial("min")])

        with comm_mode:
            z = torch.neg(d_input)

        comm_counts = comm_mode.get_total_counts()
        self.assertEqual(comm_counts, 0)
        self.assertEqual(z.placements, (Partial("max"),))

    @with_comms
    def test_maximum_mixed_partials_redistribution(self):
        # Test that mixing Partial("max") with Partial("sum") correctly
        # redistributes the incompatible partial before computing maximum
        device_mesh = self.build_device_mesh()
        comm_mode = CommDebugMode()

        input1 = torch.ones(4, 4) * (self.rank + 1)
        input2 = torch.ones(4, 4) * 0.1 * (self.rank + 1)

        d_input1 = DTensor.from_local(input1, device_mesh, [Partial("max")])
        d_input2 = DTensor.from_local(input2, device_mesh, [Partial("sum")])

        with comm_mode:
            result = torch.maximum(d_input1, d_input2)

        # Should require communication to reduce Partial("sum") to Replicate
        self.assertGreater(comm_mode.get_total_counts(), 0)
        # Result should be Partial("max") following the first operand
        self.assertEqual(result.placements, (Partial("max"),))

        # Verify correctness: d_input2's Partial("sum") should be reduced first
        # d_input2 full value = sum of all ranks' local values = 0.1 * (1+2+3+4) = 1.0
        # d_input1 stays as Partial("max"), so result.full_tensor() does max-reduce
        # max across ranks of max(rank_value, 1.0)
        # rank 0: max(1, 1) = 1, rank 1: max(2, 1) = 2, rank 2: max(3, 1) = 3, rank 3: max(4, 1) = 4
        # final max = 4
        expected_value = float(self.world_size)
        self.assertEqual(result.full_tensor()[0, 0].item(), expected_value)

    @with_comms
    def test_add_partial_with_replicate_rules(self):
        # P(x) + R -> P(x) for x in {avg, max, min}: adding a replicated constant
        # preserves the partial reduce type (constant is the same on all ranks)
        device_mesh = self.build_device_mesh()
        comm_mode = CommDebugMode()

        other_val = 3.0
        d_other = distribute_tensor(
            torch.full((8, 8), other_val, device=self.device_type),
            device_mesh,
            [Replicate()],
        )

        expected_full = {
            "avg": (self.world_size + 1) / 2.0 + other_val,
            "max": float(self.world_size) + other_val,
            "min": 1.0 + other_val,
        }

        for reduce_op in ("avg", "max", "min"):
            input_tensor = torch.ones(8, 8, device=self.device_type) * (self.rank + 1)
            d_input = DTensor.from_local(
                input_tensor, device_mesh, [Partial(reduce_op)]
            )

            with comm_mode:
                result = d_input + d_other

            self.assertEqual(comm_mode.get_total_counts(), 0)
            self.assertEqual(result.placements, (Partial(reduce_op),))
            self.assertEqual(
                result.full_tensor(),
                torch.full((8, 8), expected_full[reduce_op], device=self.device_type),
            )

    @with_comms
    def test_add_replicate_with_partial_avg(self):
        # R + P(avg) -> P(avg): avg is linear, so this holds for any alpha
        device_mesh = self.build_device_mesh()
        comm_mode = CommDebugMode()

        replicate_val = 3.0
        d_input = distribute_tensor(
            torch.full((8, 8), replicate_val, device=self.device_type),
            device_mesh,
            [Replicate()],
        )
        other_tensor = torch.ones(8, 8, device=self.device_type) * (self.rank + 1)
        d_other = DTensor.from_local(other_tensor, device_mesh, [Partial("avg")])

        with comm_mode:
            result = d_input + d_other

        self.assertEqual(comm_mode.get_total_counts(), 0)
        self.assertEqual(result.placements, (Partial("avg"),))
        # full_tensor = R_val + avg(1, 2, ..., world_size) = 3 + (ws+1)/2
        expected = torch.full(
            (8, 8),
            replicate_val + (self.world_size + 1) / 2.0,
            device=self.device_type,
        )
        self.assertEqual(result.full_tensor(), expected)

    @with_comms
    def test_mul_div_replicate_partial_asymmetry(self):
        # mul is bilinear: r * (p1 + p2) = r*p1 + r*p2, so R * P(sum) -> P(sum).
        # div is only linear in the numerator: r / (p1 + p2) != r/p1 + r/p2.
        device_mesh = self.build_device_mesh()
        comm_mode = CommDebugMode()

        replicate_val = 3.0
        partial_local_val = 2.0

        d_replicate = distribute_tensor(
            torch.full((8, 8), replicate_val, device=self.device_type),
            device_mesh,
            [Replicate()],
        )

        # R * P(sum) -> P(sum): valid, zero communication
        d_partial = DTensor.from_local(
            torch.full((8, 8), partial_local_val, device=self.device_type),
            device_mesh,
            [Partial("sum")],
        )
        with comm_mode:
            mul_result = d_replicate * d_partial

        self.assertEqual(comm_mode.get_total_counts(), 0)
        self.assertEqual(mul_result.placements, (Partial("sum"),))
        # 3 * sum(2, 2, ...) = 3 * 2 * world_size
        expected_mul = torch.full(
            (8, 8),
            replicate_val * partial_local_val * self.world_size,
            device=self.device_type,
        )
        self.assertEqual(mul_result.full_tensor(), expected_mul)

        # R / P(sum) -> requires communication (not linear in denominator)
        d_partial2 = DTensor.from_local(
            torch.full((8, 8), partial_local_val, device=self.device_type),
            device_mesh,
            [Partial("sum")],
        )
        with comm_mode:
            div_result = d_replicate / d_partial2

        self.assertGreater(comm_mode.get_total_counts(), 0)
        # After P(sum) reduction: 3 / (2 * world_size)
        expected_div = torch.full(
            (8, 8),
            replicate_val / (partial_local_val * self.world_size),
            device=self.device_type,
        )
        self.assertEqual(div_result.full_tensor(), expected_div)

    @with_comms
    def test_inplace_add_partial_avg_with_replicate(self):
        # P(avg) += R -> P(avg): self is P(avg), other is R, valid inplace
        # because the rule P(avg)+R->P(avg) keeps output == self placement.
        device_mesh = self.build_device_mesh()
        comm_mode = CommDebugMode()

        self_tensor = torch.ones(8, 8, device=self.device_type) * (self.rank + 1)
        d_self = DTensor.from_local(self_tensor, device_mesh, [Partial("avg")])
        d_other = distribute_tensor(
            torch.full((8, 8), 3.0, device=self.device_type),
            device_mesh,
            [Replicate()],
        )

        with comm_mode:
            d_self.add_(d_other)

        self.assertEqual(comm_mode.get_total_counts(), 0)
        self.assertEqual(d_self.placements, (Partial("avg"),))
        # avg(1,2,...,ws) + 3 = (ws+1)/2 + 3
        expected = torch.full(
            (8, 8),
            (self.world_size + 1) / 2.0 + 3.0,
            device=self.device_type,
        )
        self.assertEqual(d_self.full_tensor(), expected)

    @with_comms
    def test_inplace_add_replicate_with_partial_avg_requires_comm(self):
        # R += P(avg): the out-of-place rule R+P(avg)->P(avg) is valid, but
        # inplace requires output == self == R, so P(avg) output is rejected.
        # The runtime must redistribute P(avg) to R first (communication).
        device_mesh = self.build_device_mesh()
        comm_mode = CommDebugMode()

        self_tensor = torch.full((8, 8), 3.0, device=self.device_type)
        d_self = distribute_tensor(self_tensor, device_mesh, [Replicate()])
        other_tensor = torch.ones(8, 8, device=self.device_type) * (self.rank + 1)
        d_other = DTensor.from_local(other_tensor, device_mesh, [Partial("avg")])

        with comm_mode:
            d_self.add_(d_other)

        self.assertGreater(comm_mode.get_total_counts(), 0)
        self.assertEqual(d_self.placements, (Replicate(),))
        # 3 + avg(1,2,...,ws) = 3 + (ws+1)/2
        expected = torch.full(
            (8, 8),
            3.0 + (self.world_size + 1) / 2.0,
            device=self.device_type,
        )
        self.assertEqual(d_self.full_tensor(), expected)

    @with_comms
    @parametrize(
        "op_fn,second_arg,placement,expected",
        [
            # Binary list ops preserve Partial("sum")
            (torch._foreach_add, "list", Partial("sum"), Partial("sum")),
            # Scalar mul preserves Partial("sum") (linearity)
            (torch._foreach_mul, 2.0, Partial("sum"), Partial("sum")),
            # Neg preserves Partial("sum") (linearity)
            (torch._foreach_neg, None, Partial("sum"), Partial("sum")),
            # Binary list ops preserve Shard(0)
            (torch._foreach_add, "list", Shard(0), Shard(0)),
        ],
    )
    def test_foreach_placement_propagation(
        self, op_fn, second_arg, placement, expected
    ):
        device_mesh = self.build_device_mesh()
        comm_mode = CommDebugMode()

        shapes = [(8, 8), (4, 4), (2, 6)]
        if placement.is_shard():
            dts = [
                distribute_tensor(
                    torch.rand(s, device=self.device_type), device_mesh, [placement]
                )
                for s in shapes
            ]
        else:
            dts = [
                DTensor.from_local(
                    torch.rand(s, device=self.device_type), device_mesh, [placement]
                )
                for s in shapes
            ]

        if second_arg == "list":
            if placement.is_shard():
                args = [
                    distribute_tensor(
                        torch.rand(s, device=self.device_type),
                        device_mesh,
                        [placement],
                    )
                    for s in shapes
                ]
            else:
                args = [
                    DTensor.from_local(
                        torch.rand(s, device=self.device_type),
                        device_mesh,
                        [placement],
                    )
                    for s in shapes
                ]
            call_args = (dts, args)
        elif second_arg is None:
            call_args = (dts,)
        else:
            call_args = (dts, second_arg)

        with comm_mode:
            result = op_fn(*call_args)

        self.assertEqual(comm_mode.get_total_counts(), 0)
        self.assertEqual(len(result), len(shapes))
        for r in result:
            self.assertIsInstance(r, DTensor)
            self.assertEqual(r.placements, (expected,))

    @with_comms
    def test_foreach_mixed_placements(self):
        """Each element in a foreach list independently preserves its placement."""
        device_mesh = self.build_device_mesh()
        comm_mode = CommDebugMode()

        dt1 = DTensor.from_local(
            torch.rand(8, 8, device=self.device_type),
            device_mesh,
            [Partial("sum")],
        )
        ds1 = DTensor.from_local(
            torch.rand(8, 8, device=self.device_type),
            device_mesh,
            [Partial("sum")],
        )
        dt2 = distribute_tensor(
            torch.rand(8, 4, device=self.device_type), device_mesh, [Shard(0)]
        )
        ds2 = distribute_tensor(
            torch.rand(8, 4, device=self.device_type), device_mesh, [Shard(0)]
        )

        with comm_mode:
            result = torch._foreach_add([dt1, dt2], [ds1, ds2])

        self.assertEqual(comm_mode.get_total_counts(), 0)
        self.assertEqual(result[0].placements, (Partial("sum"),))
        self.assertEqual(result[1].placements, (Shard(0),))

    @with_comms
    @skip_unless_torch_gpu
    @parametrize(
        "mesh_type,amsgrad",
        [
            # All tensors on the same 1D mesh
            ("same", False),
            # state_steps on a different sub-mesh, with max_exp_avg_sqs present
            ("cross", True),
            # state_steps on a different sub-mesh, empty max_exp_avg_sqs
            # (triggers the different_mesh_args index remapping)
            ("cross", False),
        ],
    )
    def test_fused_adam(self, mesh_type, amsgrad):
        if mesh_type == "same":
            param_mesh = self.build_device_mesh()
            step_mesh = param_mesh
        else:
            mesh_2d = init_device_mesh(
                self.device_type,
                (2, self.world_size // 2),
                mesh_dim_names=("dp", "tp"),
            )
            param_mesh = mesh_2d["tp"]
            step_mesh = mesh_2d["dp"]

        d_params = [
            distribute_tensor(
                torch.rand(8, 8, device=self.device_type), param_mesh, [Shard(0)]
            )
        ]
        d_grads = [
            distribute_tensor(
                torch.rand(8, 8, device=self.device_type), param_mesh, [Shard(0)]
            )
        ]
        d_exp_avgs = [
            distribute_tensor(
                torch.zeros(8, 8, device=self.device_type), param_mesh, [Shard(0)]
            )
        ]
        d_exp_avg_sqs = [
            distribute_tensor(
                torch.zeros(8, 8, device=self.device_type), param_mesh, [Shard(0)]
            )
        ]
        d_max_exp_avg_sqs = (
            [
                distribute_tensor(
                    torch.zeros(8, 8, device=self.device_type),
                    param_mesh,
                    [Shard(0)],
                )
            ]
            if amsgrad
            else []
        )
        d_state_steps = [
            distribute_tensor(
                torch.tensor(1.0, device=self.device_type), step_mesh, [Replicate()]
            )
        ]

        torch._fused_adam_(
            d_params,
            d_grads,
            d_exp_avgs,
            d_exp_avg_sqs,
            d_max_exp_avg_sqs,
            d_state_steps,
            lr=0.001,
            beta1=0.9,
            beta2=0.999,
            weight_decay=0.0,
            eps=1e-8,
            amsgrad=amsgrad,
            maximize=False,
        )

        self.assertIsInstance(d_params[0], DTensor)
        self.assertEqual(d_params[0].placements, (Shard(0),))
        self.assertEqual(d_params[0].device_mesh, param_mesh)

    @with_comms
    @skip_unless_torch_gpu
    def test_fused_adamw_strided_shard_cross_mesh(self):
        """Reproduce the torchtitan pattern: params with _StridedShard on a 2D
        mesh, state_steps with Replicate on a 1D DP sub-mesh."""
        mesh_2d = init_device_mesh(
            self.device_type,
            (2, self.world_size // 2),
            mesh_dim_names=("dp", "tp"),
        )
        dp_mesh = mesh_2d["dp"]
        placements = [_StridedShard(0, split_factor=2), Shard(0)]

        d_params = [
            DTensor.from_local(
                torch.randn(4, 8, device=self.device_type),
                mesh_2d,
                placements,
                run_check=False,
            )
        ]
        d_grads = [
            DTensor.from_local(
                torch.randn(4, 8, device=self.device_type),
                mesh_2d,
                placements,
                run_check=False,
            )
        ]
        d_exp_avgs = [
            DTensor.from_local(
                torch.zeros(4, 8, device=self.device_type),
                mesh_2d,
                placements,
                run_check=False,
            )
        ]
        d_exp_avg_sqs = [
            DTensor.from_local(
                torch.zeros(4, 8, device=self.device_type),
                mesh_2d,
                placements,
                run_check=False,
            )
        ]
        d_state_steps = [
            DTensor.from_local(
                torch.tensor(1.0, device=self.device_type),
                dp_mesh,
                [Replicate()],
                run_check=False,
            )
        ]

        torch.ops.aten._fused_adamw_.default(
            d_params,
            d_grads,
            d_exp_avgs,
            d_exp_avg_sqs,
            [],
            d_state_steps,
            lr=0.1,
            beta1=0.9,
            beta2=0.999,
            weight_decay=0.01,
            eps=1e-8,
            amsgrad=False,
            maximize=False,
        )

        self.assertIsInstance(d_params[0], DTensor)
        self.assertEqual(d_params[0].placements, tuple(placements))
        self.assertEqual(d_params[0].device_mesh, mesh_2d)

    @with_comms
    def test_fused_adamw_tensor_lr(self):
        """Test _fused_adamw_ with tensor lr kwarg."""
        device_mesh = self.build_device_mesh()

        d_params = [
            distribute_tensor(
                torch.rand(8, 8, device=self.device_type), device_mesh, [Shard(0)]
            )
        ]
        d_grads = [
            distribute_tensor(
                torch.rand(8, 8, device=self.device_type), device_mesh, [Shard(0)]
            )
        ]
        d_exp_avgs = [
            distribute_tensor(
                torch.zeros(8, 8, device=self.device_type), device_mesh, [Shard(0)]
            )
        ]
        d_exp_avg_sqs = [
            distribute_tensor(
                torch.zeros(8, 8, device=self.device_type), device_mesh, [Shard(0)]
            )
        ]
        d_state_steps = [
            distribute_tensor(
                torch.tensor(1.0, device=self.device_type),
                device_mesh,
                [Replicate()],
            )
        ]
        d_lr = distribute_tensor(
            torch.tensor(0.001, device=self.device_type),
            device_mesh,
            [Replicate()],
        )

        # tensor_lr variant passes lr as a tensor kwarg
        torch.ops.aten._fused_adamw_.tensor_lr(
            d_params,
            d_grads,
            d_exp_avgs,
            d_exp_avg_sqs,
            [],
            d_state_steps,
            lr=d_lr,
            beta1=0.9,
            beta2=0.999,
            weight_decay=0.01,
            eps=1e-8,
            amsgrad=False,
            maximize=False,
        )

        self.assertIsInstance(d_params[0], DTensor)
        self.assertEqual(d_params[0].placements, (Shard(0),))

    @with_comms
    def test_new_pointwise_ops(self):
        device_mesh = self.build_device_mesh()
        torch.manual_seed(self.rank)
        shard_spec = [Shard(0)]

        # hardshrink: non-decreasing unary, decomp-blocked
        inp = torch.randn(8, 5, device=self.device_type)
        dt = DTensor.from_local(inp, device_mesh, shard_spec)
        self.assertEqual(
            torch.nn.functional.hardshrink(dt).to_local(),
            torch.nn.functional.hardshrink(inp),
        )

        # lcm: binary integer pointwise
        a = torch.randint(1, 100, (8, 5), device=self.device_type)
        b = torch.randint(1, 100, (8, 5), device=self.device_type)
        da = DTensor.from_local(a, device_mesh, shard_spec)
        db = DTensor.from_local(b, device_mesh, shard_spec)
        self.assertEqual(torch.lcm(da, db).to_local(), torch.lcm(a, b))

        # special_xlog1py and special_entr
        pos = torch.rand(8, 5, device=self.device_type) + 0.1
        dp = DTensor.from_local(pos, device_mesh, shard_spec)
        self.assertEqual(
            torch.special.xlog1py(dp, dp).to_local(),
            torch.special.xlog1py(pos, pos),
        )
        self.assertEqual(
            torch.special.entr(dp).to_local(),
            torch.special.entr(pos),
        )

    @with_comms
    def test_new_special_pointwise_ops(self):
        device_mesh = self.build_device_mesh()
        torch.manual_seed(self.rank)
        shard_spec = [Shard(0)]

        # Unary special functions
        inp = torch.rand(8, 5, device=self.device_type) + 0.1
        dt = DTensor.from_local(inp, device_mesh, shard_spec)
        for op in [
            torch.special.airy_ai,
            torch.special.bessel_y0,
            torch.special.bessel_y1,
            torch.special.modified_bessel_i0,
            torch.special.modified_bessel_i1,
            torch.special.modified_bessel_k0,
            torch.special.modified_bessel_k1,
            torch.special.scaled_modified_bessel_k0,
            torch.special.scaled_modified_bessel_k1,
        ]:
            self.assertEqual(op(dt).to_local(), op(inp))

        # Binary special polynomial ops
        x = torch.rand(8, 5, device=self.device_type)
        n = torch.randint(0, 5, (8, 5), device=self.device_type, dtype=x.dtype)
        dx = DTensor.from_local(x, device_mesh, shard_spec)
        dn = DTensor.from_local(n, device_mesh, shard_spec)
        for op in [
            torch.special.chebyshev_polynomial_t,
            torch.special.chebyshev_polynomial_u,
            torch.special.hermite_polynomial_h,
            torch.special.hermite_polynomial_he,
            torch.special.laguerre_polynomial_l,
            torch.special.legendre_polynomial_p,
        ]:
            self.assertEqual(op(dx, dn).to_local(), op(x, n))


instantiate_parametrized_tests(DistElementwiseOpsTest)


class TestPointwiseRuleValidation(TestCase):
    """Validate registered partial-placement rules via OpInfo samples."""

    world_size = 2

    def setUp(self):
        super().setUp()
        if not torch.distributed.is_initialized():
            torch.distributed.init_process_group(
                "fake", rank=0, world_size=self.world_size
            )

    def tearDown(self):
        super().tearDown()
        if torch.distributed.is_initialized():
            torch.distributed.destroy_process_group()

    def _with_even_sizes(self, fn):
        """Run fn with opinfo sizes overridden to be evenly divisible by world_size."""
        import torch.testing._internal.common_methods_invocations as common_ops
        from torch.testing._internal.opinfo import core as opinfo_core

        orig_sizes = (opinfo_core.L, opinfo_core.M, opinfo_core.S, opinfo_core.XS)
        opinfo_core.L = common_ops.L = 24
        opinfo_core.M = common_ops.M = 12
        opinfo_core.S = common_ops.S = 4
        opinfo_core.XS = common_ops.XS = 2
        try:
            return fn()
        finally:
            (
                opinfo_core.L,
                opinfo_core.M,
                opinfo_core.S,
                opinfo_core.XS,
            ) = orig_sizes
            (
                common_ops.L,
                common_ops.M,
                common_ops.S,
                common_ops.XS,
            ) = orig_sizes

    def test_per_category_no_incorrect_rules(self):
        """Verify no incorrect rules for one representative op per category.

        Each category of specialized pointwise ops shares the same partial
        placement rules. Validating one representative per category catches
        any mathematically incorrect rules via OpInfo samples (including
        edge cases like negative alpha for add/sub).
        """
        from torch.distributed.tensor._ops.strategy_validation import compare_operator

        representative_ops = [
            # (op_name, category, expected_true_positives)
            ("add", "binary_additive", 81),
            ("mul", "binary_mul", 49),
            ("div", "binary_div", 67),
            ("sigmoid", "non_decreasing_unary", 5),
            ("erfc", "non_increasing_unary", 5),
            ("neg", "neg", 6),
            ("maximum", "monotonic_max_preserving", 57),
            ("minimum", "monotonic_min_preserving", 57),
            ("logaddexp", "monotonic_binary", 49),
            ("abs", "pointwise_generic", 2),
        ]

        for op_name, category, expected_tp in representative_ops:
            with self.subTest(op=op_name, category=category):

                def run(name=op_name, expected=expected_tp):
                    stats = compare_operator(
                        name,
                        device="cpu",
                        dtype=torch.float32,
                        world_size=self.world_size,
                        incorrect_only=True,
                    )
                    self.assertEqual(
                        stats.true_positives,
                        expected,
                        f"{name}: expected {expected} true_positives, got {stats.true_positives}",
                    )
                    self.assertEqual(
                        len(stats.false_positives),
                        0,
                        f"{name}: found incorrect rules: {stats.false_positives}",
                    )

                self._with_even_sizes(run)


DistElementwiseOpsTestWithLocalTensor = create_local_tensor_test_class(
    DistElementwiseOpsTest, base_class=LocalDTensorOpTestBase
)


if __name__ == "__main__":
    run_tests()
